
/* (c) 2001-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <signal.h>
#include <time.h>

#include <SDL/SDL.h>
#include <SDL/SDL_video.h>

#include "defines.h"
#include "main.h"
#include "ogl.h"
#include "gui.h"
#include "ai.h"
#include "../common/parse.h"
#include "../common/socket.h"
#include "../client/oload.h"
#include "../client/world.h"
#include "../client/network.h"
#include "../client/time.h"
#include "../client/connection.h"
#include "../client/game_object.h"
#include "../client/collision.h"
#include "../client/font.h"
#include "../client/console.h"
#include "../client/txt.h"


extern struct txt_object *txt_first, *txt_last;
extern int fps;

#ifdef INGAME_DEBUG
extern int draw_tri, cube_count, object_count;
#endif

struct game_object player_object;

volatile int connection_status = OFF;
char session_passwd[256], server_name[256];
int server_port = 6789, client_port, actual_port, actual_socket;
int player_motion = 0, player_status = 0;
int today_hours = 20, today_minutes = 0, today_seconds = 0;

int status = 0;
char message[256], world_name[256];



int main(int argc, char *argv[]) {

  srand(time(NULL));
  collision_init();
  txt_init("../data/txt/");
  network_init();

  /* write default values (if connection to server fails) */
  player_object.x = 147.51;
  player_object.y = 80;
  player_object.z = 140.83;
  player_object.ax = 0;
  player_object.ay = 0;
  player_object.az = 0;
  player_object.vx = 0;
  player_object.vy = 0;
  player_object.vz = 0;
  strcpy(world_name, "2");

  /* init signal handler */
  if (signal(SIGINT, signal_handler_int) == SIG_IGN)
    return 1;
  if (signal(SIGKILL, signal_handler_int) == SIG_IGN)
    return 1;
  if (signal(SIGTERM, signal_handler_int) == SIG_IGN)
    return 1;
  if (signal(SIGSEGV, signal_handler_int) == SIG_IGN)
    return 1;
  if (signal(SIGFPE, signal_handler_int) == SIG_IGN)
    return 1;
  if (signal(SIGILL, signal_handler_int) == SIG_IGN)
    return 1;

  /* copy default settings */
  strcpy(session_passwd, "candy!");
  strcpy(server_name, "127.0.0.1");
  strcpy(player_object.name, "cherry");

  connect_to_server();

  /* no connection -> exit */
  if (connection_status == OFF) {
    fprintf(stderr, "MAIN: Could not connect to the server.\n");
    exit(1);
  }

  fprintf(stderr, "MAIN: Entering world \"%s\"...\n", world_name);

  init_window_main();

  world_init(world_name);
  game_object_init();
  font_init();
  console_init();
  object_load_list();
  ogl_init();
  ai_init();

  main_window_set_title();
  init_main_glarea();

  /* get all data in one turn */
  if (connection_status == ON) {
    send_name();
    send_visibility();
    send_flush();
    read_server_data();
  }

  SDL_ShowCursor(SDL_DISABLE);
  SDL_SetEventFilter(sdl_filter_events);

  /* GAME MAIN LOOP */
  while (1) {
    keyboard_io();
    compute_new_position();
    time_adjust_light(today_hours, today_minutes, today_seconds);
    draw_main_glarea();
    communicate_with_server();
    if ((status & STATUS_QUIT) != 0 && connection_status == OFF)
      break;
  }

  txt_free_all();

  SDL_ShowCursor(SDL_ENABLE);
  SDL_Quit();

  return 0;
}


void signal_handler_int(int signum) {

  if (connection_status == ON) {
    send_quit();
  }
  status |= STATUS_QUIT;
}


int process_client_command(char *b) {

  char t[256], tmp[256];
  int i;


  sscanf(b, "%s", t);

  /* x - close client */
  if (strcmp(t, "x") == 0) {
    send_quit();
    status |= STATUS_QUIT;
    console_write("Closing client...");
    return SUCCEEDED;
  }
  /* p - [none] - print position */
  else if (strcmp(t, "p") == 0) {
    sprintf(tmp, "(%.2f, %.2f, %.2f)", player_object.x, player_object.y, player_object.z);
    console_write(tmp);
    return SUCCEEDED;
  }
  /* fps - [none] - print fps */
  else if (strcmp(t, "fps") == 0) {
    sprintf(tmp, "fps: %d", fps);
    console_write(tmp);
    return SUCCEEDED;
  }
  /* t - [none] - print time */
  else if (strcmp(t, "t") == 0) {
    sprintf(tmp, "%.2d:%.2d:%.2d", today_hours, today_minutes, today_seconds);
    console_write(tmp);
    return SUCCEEDED;
  }
  /* s - say something */
  else if (strcmp(t, "s") == 0) {
    if (connection_status == OFF) {
      console_write("Not connected.");
      return FAILED;
    }
    strcpy(message, b);
    return SUCCEEDED;
  }
  /* cl - close connection */
  else if (strcmp(t, "cl") == 0) {
    if (connection_status == OFF) {
      console_write("Not connected.");
      return FAILED;
    }
    send_quit();
    console_write("Closing connection...");
    return SUCCEEDED;
  }
  /* co - handshake - connect to server */
  else if (strcmp(t, "co") == 0) {
    if (connection_status == ON) {
      sprintf(tmp, "Already connected to %s:%d.", server_name, actual_port);
      console_write(tmp);
      return FAILED;
    }
    sprintf(tmp, "Connecting to %s:%d", server_name, actual_port);
    console_write(tmp);
    i = connect_to_server();
    if (i == SUCCEEDED) {
      console_write("SUCCEEDED.");
      send_name();
      send_visibility();
      send_flush();
      read_server_data();
      main_window_set_title();
      return SUCCEEDED;
    }
    console_write("FAILED.");
    return FAILED;
  }
  /* csn - [none] - change server name */
  else if (strcmp(t, "csn") == 0) {
    sscanf(b, "%*s %s", t);
    if (connection_status == ON) {
      sprintf(tmp, "Already connected to %s:%d.", server_name, actual_port);
      console_write(tmp);
      return FAILED;
    }
    strcpy(server_name, t);
    sprintf(tmp, "Changed server name to \"%s\".", server_name);
    console_write(tmp);
    return SUCCEEDED;
  }
  /* e - [none] - empty console */
  else if (strcmp(t, "e") == 0) {
    console_empty();
    return SUCCEEDED;
  }
  /* cmp - [none] - change main port */
  else if (strcmp(t, "cmp") == 0) {
    sscanf(b, "%*s %d", &server_port);
    sprintf(tmp, "New main port %d.", server_port);
    console_write(tmp);
    return SUCCEEDED;
  }
  /* chp - [none] - change passwd */
  else if (strcmp(t, "chp") == 0) {
    sscanf(b, "%*s %s", t);
    if (connection_status == ON) {
      sprintf(tmp, "Already connected to %s:%d.", server_name, actual_port);
      console_write(tmp);
      return FAILED;
    }
    strcpy(session_passwd, t);
    sprintf(tmp, "New session passwd \"%s\".", session_passwd);
    console_write(tmp);
    return SUCCEEDED;
  }
#ifdef INGAME_DEBUG
  /* v - [none] - view game engine status */
  else if (strcmp(t, "v") == 0) {
    sprintf(tmp, "tris: %d cubes: %d objs: %d", draw_tri, cube_count, object_count);
    console_write(tmp);
    return SUCCEEDED;
  }
#endif
  /* cn - change name */
  else if (strcmp(t, "cn") == 0) {
    if (connection_status == OFF) {
      sscanf(b, "%*s %s", player_object.name);
      console_write("Not connected.");
      sprintf(tmp, "Your current _local_ name is \"%s\".", player_object.name);
      console_write(tmp);
      return SUCCEEDED;
    }
    else {
      sscanf(b, "%*s %s", t);
      if (strlen(t) == 0)
	return SUCCEEDED;
      sprintf(tmp, "N%s%c", t, 0);
      send_data(tmp, strlen(t) + 2);
    }
    return SUCCEEDED;
  }
  /* unknown command */
  else {
    console_write("Unknown command. Available commands:");

#ifdef INGAME_DEBUG
    console_write("cl, co, cn, csn, chp, e, fps, cmp, p, s, t, v, x");
#else
    console_write("cl, co, cn, csn, chp, e, fps, cmp, p, s, t, x");
#endif

    return FAILED;
  }

  return SUCCEEDED;
}
