
/* (c) 2001-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include <SDL/SDL_video.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "defines.h"
#include "ogl.h"
#include "gui.h"
#include "ai.h"
#include "../common/frustum.h"
#include "../common/quaternion.h"
#include "../common/euler.h"
#include "../client/oload.h"
#include "../client/world.h"
#include "../client/txt.h"


extern struct object_object **world_map;
extern struct game_object player_object;
extern struct game_object **game_object_map;
extern struct vector radius;
extern double player_x, player_y, today_light, today_fog;
extern int world_x, world_y, world_z, console_status, cross_status, console_txt, lantern_status, *object_type;
double cross_x, cross_y, cross_z;
float frustum[6*4];
int show_blocks = 5;
int frame_count = 0, frame_number = 0, previous_ticks = 0, fps = 0, cube_count = 0, object_count = 0;
int texturing_status = OFF;

/* light and colors */
GLfloat light_ambient_max[4], light_ambient[4], color_sky_max[3], color_sky[3];

struct game_object *closest_object = NULL;

#ifdef INGAME_DEBUG
int draw_tri;
#endif

struct object_object **object_array = NULL;
int object_array_size = 2048;

/* room for frame buffer (color -> object id transformation) */
unsigned char *frame_buffer = NULL;

/* list of visible objects for the ai routines, listed by the frustum culler */
struct game_object *visible_objects;



int _ogl_object_array_compare(const void *a, const void *b) {

  struct object_object *c, *d;


  c = *(struct object_object **)a;
  d = *(struct object_object **)b;
  if (c->distance < d->distance)
    return -1;
  return 1;
}


int ogl_init(void) {

  object_array = malloc(sizeof(struct object_object *) * object_array_size);
  frame_buffer = malloc(SCREEN_X*SCREEN_Y*4);
  if (object_array == NULL || frame_buffer == NULL) {
    fprintf(stderr, "OGL_INIT: Out of memory error.\n");
    return FAILED;
  }

  return SUCCEEDED;
}


int init_main_glarea(void) {

  glViewport(0, 0, SCREEN_X, SCREEN_Y);

  return SUCCEEDED;
}


int draw_main_glarea(void) {

  struct txt_object *tx;
  struct game_object *g;
  double qx, qy, qz, lx, ly, lz;
  float dx, dy, dz;
  int x, y, z, ex, ey, ez, ax, ay, az, i;
  GLint vp[4];


#ifdef INGAME_DEBUG
  /* so far 0 triangles have been drawn this frame... */
  draw_tri = 0;
#endif

  /* update world blocks */
  world_dynamic_refresh();

  /* advance "time" */
  frame_count++;
  frame_number++;

  /* player normal */
  qx = 0;
  qy = 1;
  qz = 0;
  euler_rotate(&qx, &qy, &qz, M_PI*(-player_y)/180, 0, 0);
  euler_rotate(&qx, &qy, &qz, 0, 2*M_PI-player_x/180*M_PI, 0);

  /* player looks at */
  lx = 1;
  ly = 0;
  lz = 0;
  euler_rotate(&lx, &ly, &lz, M_PI*(-player_y)/180, 0, 0);
  euler_rotate(&lx, &ly, &lz, 0, 2*M_PI-player_x/180*M_PI, 0);

  /* save the direction normal vector */
  player_object.ax = lx;
  player_object.ay = ly;
  player_object.az = lz;
  lx *= 400;
  ly *= 400;
  lz *= 400;

  lx += player_object.x;
  ly += player_object.y + radius.y/2;
  lz += player_object.z;

  /* refresh all changed textures */
  txt_refresh();

  /* opengl mode definitions */
  glClearColor(0, 0, 0, 1);
  glShadeModel(GL_FLAT);
  glDisable(GL_DITHER);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  glDepthFunc(GL_LESS);
  glFrontFace(GL_CCW);
  glEnable(GL_CULL_FACE);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  glGetIntegerv(GL_VIEWPORT, vp);

  glBlendFunc(GL_SRC_COLOR, GL_ONE);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45, (double)vp[2]/(double)vp[3], 0.1, show_blocks*CUBE_SIDE);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(player_object.x, player_object.y + radius.y/2, player_object.z, lx, ly, lz, qx, qy, qz);
  glPushMatrix();

  glEnable(GL_DEPTH_TEST);

  /* obtain the current view frustum */
  frustum_get(frustum);

  /* draw the world */
  ax = (player_object.x / CUBE_SIDE) - show_blocks;
  ay = (player_object.y / CUBE_SIDE) - show_blocks;
  az = (player_object.z / CUBE_SIDE) - show_blocks;
  ex = ax + (show_blocks<<1) + 1;
  ey = ay + (show_blocks<<1) + 1;
  ez = az + (show_blocks<<1) + 1;

  /* don't show cubes that are behind the viewer */
  z = get_player_x_segment();
  if (z == 0)
    ax += show_blocks - 2;
  else if (z == 1)
    az += show_blocks - 2;
  else if (z == 2)
    ex -= show_blocks - 1;
  else
    ez -= show_blocks - 1;

  if (ax < 0)
    ax = 0;
  if (ay < 0)
    ay = 0;
  if (az < 0)
    az = 0;
  if (ex > world_x)
    ex = world_x;
  if (ey > world_y)
    ey = world_y;
  if (ez > world_z)
    ez = world_z;

  /* build an array of all visible cubes */
  for (i = 0, z = az; z < ez; z++) {
    for (y = ay; y < ey; y++) {
      for (x = ax; x < ex; x++) {
	if (world_map[z*world_x*world_y + y*world_x + x] != NULL) {
	  if (frustum_sphere_inside(frustum, (x+0.5)*CUBE_SIDE, (y+0.5)*CUBE_SIDE, (z+0.5)*CUBE_SIDE, CUBE_SIDE) == YES) {
	    object_array[i] = world_map[z*world_x*world_y + y*world_x + x];
	    /* compute player's squared distance to the midpoint */
	    dx = (x+0.5)*CUBE_SIDE - player_object.x;
	    dy = (y+0.5)*CUBE_SIDE - player_object.y;
	    dz = (z+0.5)*CUBE_SIDE - player_object.z;
	    object_array[i]->distance = dx*dx + dy*dy + dz*dz;
	    i++;
	  }
	}
      }
    }
  }

  cube_count = i;

  /* sort the cubes by distance to the player */
  qsort(object_array, i, sizeof(struct object_object *), _ogl_object_array_compare);

  /* render the cubes from near to far */
  for (x = 0; x < i; x++) {
    object_array[x]->style = 0;
    _render_object(object_array[x]);
  }

  /* draw the game objects */
  object_count = 0;
  visible_objects = NULL;
  for (z = az; z < ez; z++) {
    for (y = ay; y < ey; y++) {
      for (x = ax; x < ex; x++) {
	g = game_object_map[z*world_x*world_y + y*world_x + x];
	while (g != NULL) {
	  glPopMatrix();
	  glPushMatrix();

	  game_object_transform(g);
	  if (render_object(g) == SUCCEEDED) {
	    /* the object was rendered, so it must be inside the view frustum */
	    g->next_ext = visible_objects;
	    g->hits = 0;
	    visible_objects = g;
	  }

	  g = g->next;
	}
      }
    }
  }

  glDisable(GL_CULL_FACE);
  glDisable(GL_DEPTH_TEST);

  /* read the frame buffer for visible object recognition */
  glReadPixels(0, 0, SCREEN_X, SCREEN_Y, GL_RGBA, GL_UNSIGNED_BYTE, frame_buffer);

  /* run the ai */
  ai_pass(frame_buffer, SCREEN_X, SCREEN_Y, visible_objects);

  /* load HUD */
  if (console_status == ON) {

    if (texturing_status == OFF)
      glEnable(GL_TEXTURE_2D);

    glBlendFunc(GL_SRC_COLOR, GL_ONE);
    glEnable(GL_BLEND);
    glColor4f(1, 1, 1, 0.5);

    txt_get(console_txt, &tx);
    glBindTexture(GL_TEXTURE_2D, tx->ogl_name);

    glLoadIdentity();
    gluLookAt(0, 0, 0, 0, 0, 400, 0, 1, 0);
    glTranslatef(0, -1, 3);
    glRotatef(180, 0, 1, 0);

    glBegin(GL_TRIANGLES);

    glTexCoord2f(0, 1);
    glVertex3f(-1, -0.1, 0);
    glTexCoord2f(1, 0);
    glVertex3f(1, 0.9, 0);
    glTexCoord2f(1, 1);
    glVertex3f(1, -0.1, 0);

    glTexCoord2f(0, 1);
    glVertex3f(-1, -0.1, 0);
    glTexCoord2f(1, 0);
    glVertex3f(1, 0.9, 0);
    glTexCoord2f(0, 0);
    glVertex3f(-1, 0.9, 0);

    glEnd();
    glDisable(GL_BLEND);

    if (texturing_status == OFF)
      glDisable(GL_TEXTURE_2D);
  }

  /* still keep the player's viewpoint for lighting */
  glPopMatrix();

  glColor4f(1, 1, 1, 1);

  /* swap buffers */
  SDL_GL_SwapBuffers();

  /* compute fps */
  x = SDL_GetTicks();
  if (x - previous_ticks > 1000) {
    fps = frame_count * 1000.0 / (x - previous_ticks);
    frame_count = 0;
    previous_ticks = x;
  }

  return SUCCEEDED;
}


int render_object(struct game_object *g) {

  struct object_object *o;
  struct object_mesh *m;
  float dx, dy, dz;


  /* infer the object's lod */
  dx = g->x - player_object.x;
  dy = g->y - player_object.y;
  dz = g->z - player_object.z;
  dx = dx*dx + dy*dy + dz*dz;

  if (dx < LOD_1_DISTANCE)
    object_get(g->object, 1, &o);
  else
    object_get(g->object, 2, &o);

  if (o == NULL)
    return FAILED;

  /* check if the object is inside the view frustum */
  m = o->meshes;
  while (m != NULL) {
    if (frustum_sphere_inside(frustum, g->x+m->bounding_sphere[0], g->y+m->bounding_sphere[1], g->z+m->bounding_sphere[2], m->bounding_sphere_r) == YES)
      break;
    m = m->next;
  }

  /* were all the parts out of view frustum? */
  if (m == NULL)
    return FAILED;

  object_count++;

  if (object_type[g->object] == OBJECT_PICKABLE) {
    o->style = g->id+1;
    return _render_object(o);
  }
  else {
    o->style = 0;
    return _render_object(o);
  }
}


int _render_object(struct object_object *o) {

  struct object_mesh *s;
  struct txt_object *tx;
  double *d;
  float *q;
  int i, *a, id, *t, h, style;
  unsigned char r, g, b;


  if (o == NULL)
    return FAILED;

  /* do we have a display list of the object? */
  if (o->display_list != 0) {
    glCallList(o->display_list);

#ifdef INGAME_DEBUG
    s = o->meshes;
    while (s != NULL) {
      draw_tri += s->triangles;
      s = s->next;
    }
#endif

    return SUCCEEDED;
  }

  style = o->style;

  if (style > 0) {
    style--;
    /* code the id into the color. when r >= 128 then the pixel
       belongs to an interesting object (id ~ color) */
    r = 128 | ((style >> 16) & 127);
    g = (style >> 8) & 255;
    b = style & 255;
    style++;
    glColor4ub(r, g, b, 255);
  }

  s = o->meshes;
  while (s != NULL) {
    i = s->triangles;
    a = s->triangle_list;
    d = s->vertice_list;
    q = s->vertex_normal_list;
    t = s->textures;

#ifdef INGAME_DEBUG
    draw_tri += i;
#endif

    glBegin(GL_TRIANGLES);
    while (i > 0) {
      if (style == 0) {
	/* txt not loaded yet? */
	if (id != *t) {
	  if (txt_get(*t, &tx) == SUCCEEDED) {
	    glEnd();
	    /* do we have translucent triangles? */
	    if (tx->status & TXT_STATUS_ALPHA)
	      glEnable(GL_BLEND);
	    else
	      glDisable(GL_BLEND);
	    glColor4ub(tx->r, tx->g, tx->b, tx->alpha*255);
	    glBegin(GL_TRIANGLES);
	    id = *t;
	  }
	}
      }
      t++;
      h = *(a++) * 3;
      glNormal3f(q[h + 0], q[h + 1], q[h + 2]);
      glVertex3f(d[h + 0], d[h + 1], d[h + 2]);
      h = *(a++) * 3;
      glNormal3f(q[h + 0], q[h + 1], q[h + 2]);
      glVertex3f(d[h + 0], d[h + 1], d[h + 2]);
      h = *(a++) * 3;
      glNormal3f(q[h + 0], q[h + 1], q[h + 2]);
      glVertex3f(d[h + 0], d[h + 1], d[h + 2]);
      i--;
    }
    glEnd();
    s = s->next;
  }

  return SUCCEEDED;
}


void game_object_transform(struct game_object *g) {

  /* rotate the object if it needs (no old rotation matrix available) rotating */
  if (g->status & GO_ROT_DIRTY) {
    g->status &= ~GO_ROT_DIRTY;
    /* build the rotation matrix */
    quaternion_rotation_matrix(g->ax, g->ay, g->az, g->rm);
    /* transform vertices if the object affects collisions */
    if (g->status & GO_COLLISION)
      game_object_transform_object(g);
  }

  glTranslatef(g->x, g->y, g->z);
  glMultMatrixf(g->rm);
}


int game_object_transform_object(struct game_object *g) {

  struct object_object *o;
  struct object_mesh *m;
  GLfloat *r;
  double *d;
  int n, i, k;


  if (g == NULL)
    return FAILED;

  object_get(g->object, 1, &o);
  if (o == NULL)
    return FAILED;

  /* has this object been transformed ever before? */
  if (g->rotated_vertices == NULL) {

    /* count the vertices */
    m = o->meshes;
    n = 0;
    while (m != NULL) {
      n += m->vertices;
      m = m->next;
    }

    /* allocate room for transformed vertices */
    g->rotated_vertices = (double *)malloc(sizeof(double) * n*3);
    if (g->rotated_vertices == NULL) {
      fprintf(stderr, "GAME_OBJECT_TRANSFORM_OBJECT: Out of memory error.\n");
      return FAILED;
    }
  }

  /* rotate and translate the vertices */
  m = o->meshes;
  r = g->rm;
  i = 0;
  while (m != NULL) {
    d = m->vertice_list;
    k = 0;
    for (n = 0; n < m->vertices; n++) {
      g->rotated_vertices[i+0] = r[0]*d[k+0] + r[1]*d[k+1] + r[2]*d[k+2] + g->x;
      g->rotated_vertices[i+1] = r[4]*d[k+0] + r[5]*d[k+1] + r[6]*d[k+2] + g->y;
      g->rotated_vertices[i+2] = r[8]*d[k+0] + r[9]*d[k+1] + r[10]*d[k+2] + g->z;
      i += 3;
      k += 3;
    }
    m = m->next;
  }

  return SUCCEEDED;
}
