
/* (c) 2002-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include "defines.h"
#include "clouds.h"
#include "txt.h"
#include "../common/frustum.h"
#include "../common/random.h"
#include "../common/angle.h"
#include "../common/billboard.h"


extern int dirty_textures;

extern struct game_object player_object;
extern struct vector player_ellipsoid;

#ifdef INGAME_DEBUG
extern int draw_tri;
#endif

float wind[] =     { 0, 1 };
float clouds_y[] = { 60, 100, 140 };
float clouds_s[] = { 0.05, 0.025, 0.01 };
float clouds_a[] = { 0.5, 0.7, 0.9 };
int   clouds_n[] = { 400, 500, 600 };
int   clouds_d[] = { 65, 65, 33 };
int clouds_per_layer[] = { 0, 0, 0 };
int cloud_textures_m[] = { 16, 16, 32 }, cloud_textures_max;
int cloud_layers = 3, clouds_max_size, clouds_max_amount = 1500, clouds_total = 0;
int clouds_retexture_at_entry = YES;
float clouds_max_distance, cloud_plane_size = 600;

struct cloud **cloud_list = NULL;
int *cloud_textures = NULL;

double *map = NULL, *res = NULL;
char *map_shadow = NULL;
int size, *map_triangles, cloud_count;



int clouds_init(void) {

  struct cloud *c;
  int i, n, p;



  /* init few variables */
  clouds_max_distance = cloud_plane_size;

  for (clouds_max_size = 0, i = 0; i < cloud_layers; i++) {
    if (clouds_d[i] > clouds_max_size)
      clouds_max_size = clouds_d[i];
  }

  for (cloud_textures_max = 0, i = 0; i < cloud_layers; i++) {
    cloud_textures_max += cloud_textures_m[i];
  }

  res = malloc(sizeof(double) * clouds_max_size*clouds_max_size);
  map = malloc(sizeof(double) * clouds_max_size*clouds_max_size);
  map_shadow = malloc(sizeof(char) * clouds_max_size*clouds_max_size);
  cloud_list = malloc(sizeof(struct cloud *) * clouds_max_amount);
  cloud_textures = malloc(sizeof(int) * cloud_textures_max);

  if (res == NULL || map == NULL || map_shadow == NULL || cloud_list == NULL || cloud_textures == NULL) {
    if (res != NULL) {
      free(res);
      res = NULL;
    }
    if (map != NULL) {
      free(map);
      map = NULL;
    }
    if (map_shadow != NULL) {
      free(map_shadow);
      map_shadow = NULL;
    }
    if (cloud_list != NULL) {
      free(cloud_list);
      cloud_list = NULL;
    }
    if (cloud_textures != NULL) {
      free(cloud_textures);
      cloud_textures = NULL;
    }

    fprintf(stderr, "CLOUDS_INIT: Out of memory error.\n");
    return FAILED;
  }

  for (i = 0; i < cloud_layers; i++)
    clouds_per_layer[i] = 0;

  for (i = 0; i < clouds_max_amount; i++)
    cloud_list[i] = NULL;

  for (i = 0; i < cloud_textures_max; i++)
    cloud_textures[i] = 0;

  /* create the textures */
  for (n = 0, i = 0; i < cloud_layers; i++) {
    for (p = 0; p < cloud_textures_m[i]; p++, n++) {
      if (txt_create(NULL, &cloud_textures[n], clouds_d[i]-1, clouds_d[i]-1) == FAILED) {
	return FAILED;
      }
      clouds_repaint_texture(cloud_textures[n], i);
    }
  }

  /* create some clouds */
  for (i = 0; i < cloud_layers; i++) {
    while (clouds_per_layer[i] < clouds_n[i]) {
      c = clouds_create(i);
      clouds_relocate_startup(c);
    }
  }

  return SUCCEEDED;
}


int clouds_manage(void) {

  static int timer = 0;
  struct cloud *c;
  float d, dx, dz;
  int i;


  /* check the amount */
  if (timer == 100) {
    timer = 0;
    for (i = 0; i < cloud_layers; i++) {
      if (clouds_per_layer[i] < clouds_n[i]) {
	/* not enough clouds around -> create one more */
	c = clouds_create(i);
	/* create coordinates */
	clouds_relocate_runtime(c);
      }
    }
  }

  timer++;

  /* move the clouds */
  for (i = 0; i < clouds_total; i++) {
    if (cloud_list[i] != NULL) {
      c = cloud_list[i];
      dx = player_object.x - c->x;
      dz = player_object.z - c->z;
      d = fabs(dx) + fabs(dz);
      if (d < clouds_max_distance) {
	if (c->alpha < clouds_a[c->layer])
	  c->alpha += 0.003;
      }
      if (d > clouds_max_distance) {
	c->alpha -= 0.003;
	if (c->alpha <= 0) {
	  clouds_relocate_runtime(c);
	  if (clouds_retexture_at_entry == YES)
	    clouds_retexture(c);
	}
      }
      clouds_move(c);
    }
  }

  return SUCCEEDED;
}


int clouds_relocate_runtime(struct cloud *c) {

  float x, z, a;


  /* relocate the clouds to where the wind blows from */
  a = angle_2d(wind[0], wind[1]) + M_PI/2;
  if (a >= M_PI*2)
    a -= M_PI*2;
  x = cos(a);
  z = sin(a);

  a = rand()/(double)RAND_MAX*cloud_plane_size - (cloud_plane_size/2);

  c->alpha = 0;
  c->x = x*a - wind[0]*cloud_plane_size/2 + player_object.x;
  c->y = clouds_y[c->layer];
  c->z = z*a - wind[1]*cloud_plane_size/2 + player_object.z;

  return SUCCEEDED;
}


int clouds_relocate_startup(struct cloud *c) {

  float x, z;


  x = rand()/(double)RAND_MAX*cloud_plane_size - (cloud_plane_size/2);
  z = rand()/(double)RAND_MAX*cloud_plane_size - (cloud_plane_size/2);

  c->alpha = 0;
  c->x = x + player_object.x;
  c->y = clouds_y[c->layer];
  c->z = z + player_object.z;

  return SUCCEEDED;
}


int clouds_move(struct cloud *c) {

  float s;


  s = clouds_s[c->layer];
  c->x += s*wind[0];
  c->z += s*wind[1];

  return SUCCEEDED;
}


int clouds_retexture(struct cloud *c) {

  int i, n;


  for (n = 0, i = 0; i < c->layer; i++)
    n += cloud_textures_m[i];

  c->txt_id = cloud_textures[n + (rand() % cloud_textures_m[c->layer])];

  return SUCCEEDED;
}


int clouds_repaint_texture(int txt, int layer) {

  struct txt_object *t;
  unsigned char *d;
  float s, m;
  int x, y;


  clouds_generate_texture(clouds_d[layer], 40, 1.4);

  txt_get(txt, &t);
  if (t == NULL)
    return FAILED;

  t->status = TXT_STATUS_ALPHA | TXT_STATUS_NO_COLLISION;
  t->dirty = YES;
  dirty_textures = YES;

  /* thickness */
  m = layer+4;

  d = (unsigned char *)t->txt;
  for (y = 0; y < t->y; y++) {
    for (x = 0; x < t->x; x++) {
      s = res[y*size + x];
      *(d++) = s;
      *(d++) = s;
      *(d++) = s;
      if (s > 20)
	s = s*m;
      else
	s = s/m;
      if (s > 255)
	s = 255;
      *(d++) = s;
    }
  }

  txt_refresh();

  return SUCCEEDED;
}


struct cloud *clouds_create(int layer) {

  struct cloud *c;
  int i;


  c = malloc(sizeof(struct cloud));
  if (c == NULL) {
    fprintf(stderr, "CLOUDS_CREATE: Out of memory error.\n");
    return NULL;
  }

  c->layer = layer;
  c->scale = rand()/(double)RAND_MAX*15+5;

  /* select a new cloud texture */
  clouds_retexture(c);

  clouds_per_layer[layer]++;
  clouds_total++;

  for (i = 0; i < clouds_max_amount; i++) {
    if (cloud_list[i] == NULL) {
      cloud_list[i] = c;
      break;
    }
  }

  return c;
}


int _clouds_sort(const void *a, const void *b) {

  struct cloud *c = (struct cloud *)a, *d = (struct cloud *)b;


  if (c->layer < d->layer)
    return 1;
  if (c->layer > d->layer)
    return -1;
  if (c->d > d->d)
    return -1;
  return 1;
}


int clouds_draw_all(void) {

  GLfloat c[4*3], bbv[3*4], sc, texcoords[4*2];
  GLint bl, li, te, dm, bd, bs;
  struct txt_object *t;
  struct cloud *g;
  float dx, dy, dz;
  float frustum[6*4];
  int i;


  /* compute the distance to the player */
  for (i = 0; i < clouds_total; i++) {
    if (cloud_list[i] != NULL) {
      g = cloud_list[i];
      dx = player_object.x - g->x;
      dy = player_object.y + player_ellipsoid.y/2 - g->y;
      dz = player_object.z - g->z;
      g->d = dx*dx + dy*dy + dz*dz;
    }
  }

  /* sort the clouds (distance to the camera) */
  qsort(cloud_list, clouds_total, sizeof(struct cloud *), _clouds_sort);

  /* obtain the current view frustum */
  frustum_get(frustum);

  billboard_spherical_cheat_matrix(bbv);

  glGetIntegerv(GL_BLEND, &bl);
  glGetIntegerv(GL_LIGHTING, &li);
  glGetIntegerv(GL_TEXTURE_2D, &te);
  glGetIntegerv(GL_DEPTH_WRITEMASK, &dm);
  glGetIntegerv(GL_BLEND_DST, &bd);
  glGetIntegerv(GL_BLEND_SRC, &bs);

  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);
  glEnable(GL_TEXTURE_2D);
  glDisable(GL_LIGHTING);
  glDepthMask(GL_FALSE);

  texcoords[0*2 + 0] = 1;
  texcoords[0*2 + 1] = 1;
  texcoords[1*2 + 0] = 0;
  texcoords[1*2 + 1] = 1;
  texcoords[2*2 + 0] = 0;
  texcoords[2*2 + 1] = 0;
  texcoords[3*2 + 0] = 1;
  texcoords[3*2 + 1] = 0;

  glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  glEnableClientState(GL_VERTEX_ARRAY);

  glTexCoordPointer(2, GL_FLOAT, 0, texcoords);
  glVertexPointer(3, GL_FLOAT, 0, c);

  cloud_count = 0;
  for (i = 0; i < clouds_total; i++) {
    if (cloud_list[i] != NULL) {
      g = cloud_list[i];

      /* can we see it? */
      if (frustum_sphere_inside(frustum, g->x, g->y, g->z, g->scale) == NO)
	continue;

      cloud_count++;
      sc = g->scale;
      c[2*3 + 0] = g->x + bbv[0]*sc;
      c[2*3 + 1] = g->y + bbv[1]*sc;
      c[2*3 + 2] = g->z + bbv[2]*sc;
      c[3*3 + 0] = g->x + bbv[3]*sc;
      c[3*3 + 1] = g->y + bbv[4]*sc;
      c[3*3 + 2] = g->z + bbv[5]*sc;
      c[0*3 + 0] = g->x + bbv[6]*sc;
      c[0*3 + 1] = g->y + bbv[7]*sc;
      c[0*3 + 2] = g->z + bbv[8]*sc;
      c[1*3 + 0] = g->x + bbv[9]*sc;
      c[1*3 + 1] = g->y + bbv[10]*sc;
      c[1*3 + 2] = g->z + bbv[11]*sc;

      /* draw the cloud */
      txt_get(g->txt_id, &t);

      /* compute the color */
      if (g->layer == 0)
	dx = 0.4;
      else if (g->layer == 1)
	dx = 0.3;
      else
	dx = 0.2;

      dx = dx*(clouds_max_distance*clouds_max_distance/4 - g->d)/clouds_max_distance/clouds_max_distance*4;
      if (dx < 0)
	dx = 0;

      glColor4f(dx/1.3, dx/1.2, dx, g->alpha*t->alpha);

      glBindTexture(GL_TEXTURE_2D, t->ogl_name);

      glDrawArrays(GL_QUADS, 0, 4);

#ifdef INGAME_DEBUG
      draw_tri += 2;
#endif
    }
  }

  /*
  glDisableClientState(GL_VERTEX_ARRAY);
  glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  */

  /* return old opengl status */
  glDepthMask(dm);
  glBlendFunc(bs, bd);
  if (bl == GL_FALSE)
    glDisable(GL_BLEND);
  if (li == GL_TRUE)
    glEnable(GL_LIGHTING);
  if (te == GL_FALSE)
    glDisable(GL_TEXTURE_2D);

  return SUCCEEDED;
}


static int _clouds_map_insert(int x, int y, double d) {

  if (x < 0 || x >= size)
    return FAILED;
  if (y < 0 || y >= size)
    return FAILED;

  if (d < 0)
    d = 0;

  if (map_shadow[y * size + x] == 0) {
    map_shadow[y * size + x] = 1;
    map[y * size + x] = d;
    return SUCCEEDED;
  }

  return FAILED;
}


static int _clouds_map_read(int x, int y, double *d) {

  while (x < 0)
    x += size;
  while (y < 0)
    y += size;
  while (x >= size)
    x -= size;
  while (y >= size)
    y -= size;

  *d += map[y * size + x];
  return SUCCEEDED;
}


static int _clouds_smooth_texture(void) {

  double d;
  int x, y, i;


  for (i = 0, y = 0; y < size; y++) {
    for (x = 0; x < size; x++, i++) {
      d = 0;
      _clouds_map_read(x + 1, y - 1, &d);
      _clouds_map_read(x + 1, y + 0, &d);
      _clouds_map_read(x + 1, y + 1, &d);
      _clouds_map_read(x - 1, y - 1, &d);
      _clouds_map_read(x - 1, y + 0, &d);
      _clouds_map_read(x - 1, y + 1, &d);
      _clouds_map_read(x + 0, y - 1, &d);
      _clouds_map_read(x + 0, y + 0, &d);
      _clouds_map_read(x + 0, y + 1, &d);
      res[i] = d/9;
    }
  }

  return SUCCEEDED;
}


int clouds_generate_texture(int siz, double limit, double divider) {

  double d, t;
  int r, x = 1, y = 1, a, b, g, q, p;


  size = siz;
  r = size - 1;

  for (p = 0; p < size*size; p++) {
    map[p] = 0;
    map_shadow[p] = 0;
  }

  /* init borders */
  for (a = 0; a < size; a++) {
    _clouds_map_insert(a, 0, 0);
    _clouds_map_insert(0, a, 0);
    _clouds_map_insert(a, size - 1, 0);
    _clouds_map_insert(size - 1, a, 0);
  }

  /* first step */
  d = map[0] + map[size - 1] + map[size * (size - 1)] + map[size * (size - 1) + size - 1];
  d /= 4;
  d += random_gaussian(limit);
  _clouds_map_insert((size - 1) / 2, (size - 1) / 2, d);

  x = x<<1;
  y = y<<1;

  /* generate a cloud using diamond-square -algorithm
     (see eg. http://www.gameprogrammer.com/fractal.html) */
  while (r > 1) {
    /* diamond steps */
    g = r>>1;
    for (q = 0, b = 0; b < y; b++, q += r) {
      for (p = 0, a = 0; a < x; a++, p += r) {
	/* right */
	d = 0;
	_clouds_map_read(p + 0, q + 0, &d);
	_clouds_map_read(p + r, q + 0, &d);
	_clouds_map_read(p + g, q + g, &d);
	_clouds_map_read(p + g, q - g, &d);
	d /= 4;
	d += random_gaussian(limit);
	_clouds_map_insert(p + g, q, d);
	/* down */
	d = 0;
	_clouds_map_read(p + 0, q + 0, &d);
	_clouds_map_read(p + 0, q + r, &d);
	_clouds_map_read(p + g, q + g, &d);
	_clouds_map_read(p - g, q + g, &d);
	d /= 4;
	d += random_gaussian(limit);
	_clouds_map_insert(p, q + g, d);
      }
    }

    r = r>>1;

    if (r <= 1)
      break;

    /* square steps */
    g = r>>1;
    for (q = 0, b = 0; b < y; b++, q += r) {
      for (p = 0, a = 0; a < x; a++, p += r) {
	d = 0;
	_clouds_map_read(p + 0, q + 0, &d);
	_clouds_map_read(p + r, q + 0, &d);
	_clouds_map_read(p + r, q + r, &d);
	_clouds_map_read(p + 0, q + r, &d);
	d /= 4;
	d += random_gaussian(limit);
	_clouds_map_insert(p + g, q + g, d);
      }
    }

    x = x<<1;
    y = y<<1;

    /* diminish the effect of the random generator as iterations go on */
    limit /= divider;
  }

  /* bring the negative bottom floor to zero level */
  for (t = map[0], d = map[0], x = 1; x < size*size; x++) {
    if (map[x] < t)
      t = map[x];
    if (map[x] > d)
      d = map[x];
  }

  d -= t;

  /* bottom to zero, and scale it up */
  for (x = 0; x < size*size; x++) {
    map[x] = (map[x]-t)*255/d;
  }

  _clouds_smooth_texture();

  return SUCCEEDED;
}
