
/* (c) 2002-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>

#include "defines.h"
#include "gui.h"
#include "network.h"
#include "console.h"
#include "connection.h"
#include "game_object.h"
#include "action.h"
#include "particles.h"
#include "../common/parse.h"
#include "../common/dynmem.h"
#include "../common/socket.h"

/* 1 = normal, 3 = all */
/*
#define NETWORK_DEBUG 3
*/

extern struct game_object *closest_object;
extern struct game_object player_object;
extern volatile int connection_status;
extern int today_hours, today_minutes, today_seconds, *object_type;
extern int actual_socket, status, player_status, show_blocks, lantern_status;

char *buffer_out = NULL, *buffer_in = NULL;
int buffer_out_size = 32*1024, buffer_out_i = 0, buffer_in_size = 128;



int network_init(void) {

  buffer_out = malloc(buffer_out_size);
  buffer_in = malloc(buffer_in_size);
  buffer_out_i = 5;

  if (buffer_out == NULL || buffer_in == NULL) {
    fprintf(stderr, "NETWORK_INIT: Out of memory error.\n");
    return FAILED;
  }

  return SUCCEEDED;
}


void send_data(char *d, int i) {

  if (connection_status == OFF)
    return;

  while (i > 0) {
    if (buffer_out_i >= buffer_out_size) {
      if (buffer_out_size == 0)
	exit(0);
      /* double the buffer size */
      buffer_out_size *= 2;
      buffer_out = realloc(buffer_out, buffer_out_size);
      if (buffer_out == NULL) {
	fprintf(stderr, "SEND_DATA: Out of memory error.\n");
	buffer_out_i = 0;
	buffer_out_size = 0;
	return;
      }
    }
    buffer_out[buffer_out_i++] = *(d++);
    i--;
  }
}


void send_flush(void) {

  if (connection_status == OFF)
    return;

  if (buffer_out_i == 5)
    return;

  /* message block size */
  buffer_out[0] = 's';
  write_int(buffer_out + 1, buffer_out_i - 5);

  send(actual_socket, buffer_out, buffer_out_i, 0);

#ifdef NETWORK_DEBUG
  fprintf(stderr, "SEND_FLUSH: Sent %d byte(s).\n", buffer_out_i);
#endif

  buffer_out_i = 5;
}


void send_visibility(void) {

  char tmp[4];


  if (connection_status == OFF)
    return;

  tmp[0] = 'V';
  tmp[1] = show_blocks;
  send_data(tmp, 2);
}


void send_lantern(void) {

  char tmp[4];


  if (connection_status == OFF)
    return;

  tmp[0] = 'L';
  tmp[1] = lantern_status;
  send_data(tmp, 2);
}


void send_name(void) {

  char tmp[256];


  if (connection_status == OFF) {
    sprintf(tmp, "Not connected. Your current _local_ name is \"%s\".", player_object.name);
    console_write(tmp);
    return;
  }

  sprintf(tmp, "N%s%c", player_object.name, 0);
  send_data(tmp, strlen(player_object.name) + 2);
}


void send_quit(void) {

  char tmp[4];


  if (connection_status == OFF)
    return;

  tmp[0] = 'D';
  send_data(tmp, 1);
}

void send_position(void) {

  char b[128];
  int x;


  if (connection_status == OFF) {
    fprintf(stderr, "SEND_POSITION: Failure. Not connected.\n");
    return;
  }

  b[0] = 'p';
  x = 1;
  write_double_a(b, &x, player_object.x);
  write_double_a(b, &x, player_object.y);
  write_double_a(b, &x, player_object.z);
  write_float_a(b, &x, player_object.ax);
  write_float_a(b, &x, player_object.ay);
  write_float_a(b, &x, player_object.az);
  send_data(b, x);
}


void communicate_with_server(void) {

  struct timeval t;
  fd_set read_fd_set;
  char tmp[256];
  int i;


  if (connection_status == OFF)
    return;

  /* has the player position/orientation changed since last frame? */
  if ((player_status & PLAYER_STATUS_MOVED) != 0) {
    player_status &= ~PLAYER_STATUS_MOVED;
    tmp[0] = 'O';
    write_int(tmp + 1, player_object.id);
    i = 5;
    write_double_a(tmp, &i, player_object.x);
    write_double_a(tmp, &i, player_object.y);
    write_double_a(tmp, &i, player_object.z);
    write_float_a(tmp, &i, player_object.ax);
    write_float_a(tmp, &i, player_object.ay);
    write_float_a(tmp, &i, player_object.az);
    send_data(tmp, i);
  }

  t.tv_sec = 0;
  t.tv_usec = 0;
  FD_ZERO(&read_fd_set);
  FD_SET(actual_socket, &read_fd_set);
  i = select(FD_SETSIZE, &read_fd_set, NULL, NULL, &t);
  if (i < 0)
    return;
  /* there's data from the server! */
  if (i > 0)
    read_server_data();

  /* send my data */
  send_flush();
}


void get_next_symbol(char *c, char *o, int *i) {

  while (c[*i] == ' ')
    (*i)++;

  while (c[*i] != ' ' && c[*i] != 0) {
    *(o++) = c[*i];
    (*i)++;
  }
  *o = 0;
}


void read_server_data(void) {

  struct particle_system *ps;
  struct game_object *g;
  char tmp[256], b[256];
  double px, py, pz;
  int s, x, d, size, c, id, object;
  float ax, ay, az;
  FILE *f;


  s = recv(actual_socket, buffer_in, 5, 0);

  /* get at least five bytes (to get the message size) */
  while (s < 5)
    s += recv(actual_socket, buffer_in + s, 5 - s, 0);

  /* parse the size */
  if (buffer_in[0] != 's') {
    fprintf(stderr, "READ_SERVER_DATA: Unknown command '%c', should be 's'.\n", buffer_in[0]);
    exit(0);
  }

  read_int(buffer_in + 1, &size);

  if (size > buffer_in_size) {
    fprintf(stderr, "READ_SERVER_DATA: Data overflow from the server (got a %d+5 byte message, room for only %d bytes).\n", size, buffer_in_size);
    fprintf(stderr, "READ_SERVER_DATA: Enlarging buffers.\n");
    dynmem_handle_buffer(&buffer_in, &buffer_in_size, size);
  }

  s = 0;
  /* get the whole message (and nothing extra) */
  while (s < size)
    s += recv(actual_socket, buffer_in + s, size - s, 0);

#ifdef NETWORK_DEBUG
  fprintf(stderr, "READ_SERVER_DATA: Got %d bytes.\n", s);
#endif

  x = 0;
  while (x < s) {
#ifdef NETWORK_DEBUG
    fprintf(stderr, "READ_SERVER_DATA: Next command: '%c' (%d). x=%d s=%d\n", buffer_in[x], buffer_in[x], x, s);
#endif
    switch (buffer_in[x]) {

    /* the death of the server */
    case 'D':
      console_write("Server closed connection.");
      socket_delete(actual_socket);
      connection_status = OFF;
      game_object_remove_all();
      main_window_set_title();
      return;

    /* someone said something */
    case 'S':
      x++;
      get_next_symbol(buffer_in, b, &x);
      sprintf(tmp, "%s said:", b);
      console_write(tmp);
      sscanf(buffer_in + x, "%d", &d);
      get_next_symbol(buffer_in, tmp, &x);
      x++;
      strncpy(b, buffer_in + x, d);
      b[d] = 0;
      sprintf(tmp, "\"%s\"", b);
      console_write(tmp);
      x += d;
      continue;

    /* time */
    case 'T':
      today_hours = buffer_in[x + 1];
      today_minutes = buffer_in[x + 2];
      today_seconds = buffer_in[x + 3];
      x += 4;
#ifdef NETWORK_DEBUG
      fprintf(stderr, "NEW TIME: %dh %dm %ds.\n", today_hours, today_minutes, today_seconds);
#endif
      continue;

    /* add to inventory */
    case 'I':
      x++;
      read_int_a(buffer_in, &x, &d);
#ifdef NETWORK_DEBUG
      fprintf(stderr, "ADD_TO_INVENTORY: Object %d.\n", d);
#endif
      action_add_to_inventory(d);
      continue;

    /* new dummy object */
    case 'd':
      x++;
      read_int_a(buffer_in, &x, &id);
#ifdef NETWORK_DEBUG
      fprintf(stderr, "NEW_DUMMY_ID: %d.\n", id);
#endif
      game_object_dummy_add(id);
      continue;

    /* new particle system */
    case 'n':
      x++;
      read_int_a(buffer_in, &x, &id);
      read_int_a(buffer_in, &x, &object);
      read_double_a(buffer_in, &x, &px);
      read_double_a(buffer_in, &x, &py);
      read_double_a(buffer_in, &x, &pz);
      read_float_a(buffer_in, &x, &ax);
      read_float_a(buffer_in, &x, &ay);
      read_float_a(buffer_in, &x, &az);

#ifdef NETWORK_DEBUG
      fprintf(stderr, "NEW_PARTICLE_SYSTEM: %d.\n", id);
#endif

      particle_system_add(id, object, px, py, pz, ax, ay, az);

      continue;

    /* new object */
    case 'N':
      x++;
      read_int_a(buffer_in, &x, &id);
      read_int_a(buffer_in, &x, &object);
      read_double_a(buffer_in, &x, &px);
      read_double_a(buffer_in, &x, &py);
      read_double_a(buffer_in, &x, &pz);
      read_float_a(buffer_in, &x, &ax);
      read_float_a(buffer_in, &x, &ay);
      read_float_a(buffer_in, &x, &az);
      strcpy(tmp, buffer_in + x);
      x += strlen(tmp) + 1;
      c = buffer_in[x++];

      /* the player will always skip himself */
      if (id == player_object.id)
	continue;

      if (game_object_create(tmp, id, object, px, py, pz, ax, ay, az, &g, c) == FAILED)
	return;

#ifdef NETWORK_DEBUG
      fprintf(stderr, "NEW_OBJECT: %d \"%s\".\n", g->id, g->name);
#endif
      continue;

    /* particle system info (updated) */
    case 'p':
      x++;
      read_int_a(buffer_in, &x, &d);

      read_double_a(buffer_in, &x, &px);
      read_double_a(buffer_in, &x, &py);
      read_double_a(buffer_in, &x, &pz);
      read_float_a(buffer_in, &x, &ax);
      read_float_a(buffer_in, &x, &ay);
      read_float_a(buffer_in, &x, &az);

      if (particle_system_get(d, &ps) == FAILED) {
	/* don't know that particle system -> request it from the server */
	tmp[0] = 'm';
	write_int(tmp + 1, d);
	send_data(tmp, 5);

#ifdef NETWORK_DEBUG
	fprintf(stderr, "UNKNOWN PARTICLE SYSTEM: id=%d. Have to ask...\n", d);
#endif

	continue;
      }

#ifdef NETWORK_DEBUG
      fprintf(stderr, "PARTICLE_SYSTEM_INFO: ID: %d POS: (%f, %f, %f) ROT: (%f, %f, %f).\n", d, px, py, pz, ax, ay, az);
#endif

      ps->x = px;
      ps->y = py;
      ps->z = pz;
      ps->dx = ax;
      ps->dy = ay;
      ps->dz = az;

      particle_system_relocate(ps);

      continue;

    /* object info (updated) */
    case 'O':
      x++;
      read_int_a(buffer_in, &x, &d);

      /* the player will always skip himself */
      if (d == player_object.id) {
	x += sizeof(double)*3 + sizeof(float)*3;
	continue;
      }

      if (game_object_get(d, &g) == FAILED) {
	/* i don't know that object -> request it from the server */
	tmp[0] = 'M';
	write_int(tmp + 1, d);
	send_data(tmp, 5);
	x += sizeof(double)*3 + sizeof(float)*3;

#ifdef NETWORK_DEBUG
	fprintf(stderr, "UNKNOWN OBJECT: id=%d. Have to ask...\n", d);
#endif

	continue;
      }

      read_double_a(buffer_in, &x, &(g->x));
      read_double_a(buffer_in, &x, &(g->y));
      read_double_a(buffer_in, &x, &(g->z));
      read_float_a(buffer_in, &x, &ax);
      read_float_a(buffer_in, &x, &ay);
      read_float_a(buffer_in, &x, &az);
      if (ax != g->ax || ay != g->ay || az != g->az) {
	/* mark it to be rotated */
	g->status |= GO_ROT_DIRTY;
	g->ax = ax;
	g->ay = ay;
	g->az = az;
      }

#ifdef NETWORK_DEBUG
      fprintf(stderr, "NEW ORIENTATION: id=%d.\n", d);
#endif

      game_object_relocate(g);

      continue;

    /* new id and position (for the player) */
    case 'P':
      x++;
      g = &player_object;
      read_int_a(buffer_in, &x, &(g->id));
      read_double_a(buffer_in, &x, &(g->x));
      read_double_a(buffer_in, &x, &(g->y));
      read_double_a(buffer_in, &x, &(g->z));
#ifdef NETWORK_DEBUG
      fprintf(stderr, "NEW_PLAYER_DATA: ID = %d, position (%f, %f, %f).\n", g->id, g->x, g->y, g->z);
#endif
      continue;

    /* delete object */
    case 'X':
      x++;
      read_int_a(buffer_in, &x, &d);
      if (game_object_get(d, &g) == FAILED)
	return;
#ifdef NETWORK_DEBUG
      fprintf(stderr, "DELETE_OBJECT: ID %d.\n", d);
#endif
      if (object_type[g->object] == OBJECT_PLAYER) {
	sprintf(tmp, "Player \"%s\" has quit.\n", g->name);
	console_write(tmp);
      }
      /* is player highlighting the object? */
      if (closest_object == g)
	closest_object = NULL;
      game_object_remove(d);
      continue;

    /* object name change */
    case 'Y':
      x++;
      read_int_a(buffer_in, &x, &d);
      game_object_get(d, &g);
      get_next_symbol(buffer_in, tmp, &x);
      if (g != NULL && object_type[g->object] == OBJECT_PLAYER) {
	sprintf(b, "Player \"%s\" changed name to \"%s\".", g->name, tmp);
	console_write(b);
      }
      if (g != NULL)
	strcpy(g->name, tmp);
      x++;
#ifdef NETWORK_DEBUG
      fprintf(stderr, "NEW_NAME: \"%s\".\n", tmp);
#endif
      continue;

    default:
      fprintf(stderr, "READ_SERVER_DATA: Unknown command '%d'.\n", buffer_in[x]);
      x++;
      /* save the buffer for debugging */
      f = fopen("in.dat", "wb");
      fwrite(buffer_in, 1, s, f);
      fflush(f);
      fclose(f);
      exit(0);
      break;
    }
  }
}
