
/* (c) 2003 ville helin <vhelin@cc.hut.fi> */

/* depth-of-field for any opengl application. the nearer to the
   camera the clearer the view. note that the current algorithm
   is really slow even on a humble geforce4 ti4600 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include "defines.h"
#include "dof.h"


unsigned char *dof_blur = NULL, *dof_sharp = NULL;
GLfloat *dof_depth = NULL;
int dof_dx = 0, dof_dy = 0;


int dof_init(void) {

  GLint vp[4];


  glGetIntegerv(GL_VIEWPORT, vp);

  /* do we need a new texture? */
  if (dof_depth == NULL || dof_blur == NULL || dof_sharp == NULL || vp[2] != dof_dx || vp[3] != dof_dy) {

    if (dof_blur != NULL)
      free(dof_blur);
    if (dof_sharp != NULL)
      free(dof_sharp);
    if (dof_depth != NULL)
      free(dof_depth);

    dof_blur = malloc(vp[2]*vp[3]*4);
    dof_sharp = malloc(vp[2]*vp[3]*4);
    dof_depth = malloc(vp[2]*vp[3]*sizeof(GLfloat));
    if (dof_blur == NULL || dof_sharp == NULL || dof_depth == NULL) {
      fprintf(stderr, "DOF_INIT: Out of memory error.\n");
      dof_exit();
      return FAILED;
    }

    dof_dx = vp[2];
    dof_dy = vp[3];
  }

  return SUCCEEDED;
}


int dof_exit(void) {

  if (dof_blur != NULL)
    free(dof_blur);
  if (dof_sharp != NULL)
    free(dof_sharp);
  if (dof_depth != NULL)
    free(dof_depth);

  return SUCCEEDED;
}


int dof_render(void) {

  GLuint tn;
  GLint vp[4];
  int x, y, i;
  float min, max, d;


  glGetIntegerv(GL_VIEWPORT, vp);

  if (dof_blur == NULL || dof_sharp == NULL || dof_depth == NULL)
    return FAILED;

  glPushMatrix();

  /* read in the sharp data */
  glReadPixels(0, 0, vp[2], vp[3], GL_RGBA, GL_UNSIGNED_BYTE, dof_sharp);

  /* read in the depth buffer */
  glReadPixels(0, 0, vp[2], vp[3], GL_DEPTH_COMPONENT, GL_FLOAT, dof_depth);

  /* normalize depth buffer as opengl doesn't seem to do it though it should */
  min = dof_depth[0];
  max = dof_depth[0];
  for (i = 0; i < vp[2]*vp[3]; i++) {
    if (dof_depth[i] < min)
      min = dof_depth[i];
    if (dof_depth[i] > max)
      max = dof_depth[i];
  }

  for (i = 0; i < vp[2]*vp[3]; i++) {
    dof_depth[i] = (dof_depth[i] - min)*(1/(max - min));
  }

  /* create the blurred texture (just a 50% shrink of the original) */
  for (i = 0, y = 0; y < vp[3]/2; y++) {
    for (x = 0; x < vp[2]/2; x++) {
      dof_blur[i + 0] = dof_sharp[(y*vp[2]*2 + x*2)*4 + 0];
      dof_blur[i + 1] = dof_sharp[(y*vp[2]*2 + x*2)*4 + 1];
      dof_blur[i + 2] = dof_sharp[(y*vp[2]*2 + x*2)*4 + 2];
      /* compute the alpha value */
      d = (unsigned char)(dof_depth[y*vp[2]*2 + x*2]*128.0);
      dof_blur[i + 3] = (unsigned char)d;
      i += 4;
    }
  }

  /* render the blurred texture */
  glDisable(GL_DEPTH_TEST);
  glDisable(GL_FOG);
  glDisable(GL_CULL_FACE);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-1, 1, -1, 1, 0, 0);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);

  glGenTextures(1, &tn);
  glBindTexture(GL_TEXTURE_2D, tn);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, vp[2]/2, vp[3]/2, 0, GL_RGBA, GL_UNSIGNED_BYTE, dof_blur);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

  glDepthMask(GL_FALSE);
  glColor4f(1, 1, 1, 1);

  glBegin(GL_QUADS);

  glTexCoord2f(0, 0);
  glVertex3f(-1, -1, 0);
  glTexCoord2f(0, 1);
  glVertex3f(-1, 1, 0);
  glTexCoord2f(1, 1);
  glVertex3f(1, 1, 0);
  glTexCoord2f(1, 0);
  glVertex3f(1, -1, 0);

  glEnd();

  glDepthMask(GL_TRUE);

  glDeleteTextures(1, &tn);

  glPopMatrix();

  return SUCCEEDED;
}
