
/* (c) 2002-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "defines.h"
#include "main.h"


unsigned char *floor_plan, room, area[256], *floor_mask, *tmp;
unsigned char *walls_hor, *walls_ver, cluster[256], *ceiling, *floor, *staircase_base;
struct neighbour *room_neighbours[256];
struct house_floor *floors = NULL;
int floor_dx, floor_dy, clusters, rooms;

char landir[] = "houseparts";

/* house parameters */
double house_x = 0, house_y = 0, house_z = 0;
int house_max_dx = 0, house_max_dy = 0;



int main(int argc, char *argv[]) {

  unsigned char *n;
  unsigned int s, i, f;


  if (argc != 7) {
    fprintf(stderr, "USAGE: %s [seed] [center x] [center y] [center z] [max dx] [max dy]\n", argv[0]);
    return 1;
  }

  s = atoi(argv[1]);
  house_x = atof(argv[2]);
  house_y = atof(argv[3]);
  house_z = atof(argv[4]);
  house_max_dx = atoi(argv[5]);
  house_max_dy = atoi(argv[6]);

  srand(s);

  floor_dy = (rand() % (house_max_dy-HOUSE_MIN_DY+1)) + HOUSE_MIN_DY;
  floor_dx = (rand() % (house_max_dx-HOUSE_MIN_DX+1)) + HOUSE_MIN_DX;

  tmp = malloc(floor_dx*floor_dy);
  if (tmp == NULL) {
    fprintf(stderr, "MAIN: Out of memory error.\n");
    return FAILED;
  }

  /* allocate space for floors and walls */
  if (houseb_create_floor() == FAILED)
    return 1;

  /* create n floors upwards */
  for (f = 0; f < 2; f++) {

    printf("**** FLOOR %d ****\n", f+1);

    room = 'a';

    /* add rooms until it's all filled up */
    while (houseb_add_room() == SUCCEEDED);

    /*
      houseb_fill_holes();
      houseb_fix_anomalities();
    */

    houseb_drop_random_rooms();
    houseb_apply_mask();
    houseb_fix_labels();
    houseb_fix_clusters();
    houseb_compute_room_areas();
    houseb_compute_neighbours();
    houseb_create_walls();
    houseb_add_doors();
    houseb_compute_ceiling();
    houseb_compute_floor();
    houseb_compute_windowed_walls();
    houseb_compute_staircase_base();
    houseb_add_floor();

    /*
      houseb_print_neighbours();
      houseb_print_floor();
      houseb_print_ceiling();
      houseb_print_walls();
      houseb_print_staircase_base();
    */
    houseb_print_floor_plan();

    /* allocate space for floors and walls */
    n = floor_plan;
    if (houseb_create_floor() == FAILED)
      return 1;

    /* floor plan mask for the next floor */
    for (i = 0; i < floor_dx*floor_dy; i++) {
      if (n[i] != '.')
	floor_mask[i] = 1;
      else
	floor_mask[i] = 0;
    }
  }

  if (houseb_place_staircases() == FAILED)
    return -1;

  houseb_output_house();

  free(tmp);

  return 0;
}


int houseb_add_floor(void) {

  struct house_floor *h, *p;
  int i;


  h = malloc(sizeof(struct house_floor));
  if (h == NULL) {
    fprintf(stderr, "HOUSEB_ADD_FLOOR: Out of memory error.\n");
    return FAILED;
  }

  h->floor_plan = (unsigned char *)malloc(floor_dx * floor_dy);
  h->floor_mask = (unsigned char *)malloc(floor_dx * floor_dy);
  h->ceiling = (unsigned char *)malloc(floor_dx * floor_dy);
  h->floor = (unsigned char *)malloc(floor_dx * floor_dy);
  h->staircase_base = (unsigned char *)malloc(floor_dx * floor_dy);
  h->walls_hor = (unsigned char *)malloc(floor_dx * floor_dy * 2);
  h->walls_ver = (unsigned char *)malloc(floor_dx * floor_dy * 2);
  if (h->floor_plan == NULL || h->walls_hor == NULL || h->walls_ver == NULL || h->floor_mask == NULL) {
    fprintf(stderr, "HOUSEB_ADD_FLOOR: Out of memory error.\n");
    return FAILED;
  }

  /* copy all data to the structure */
  for (i = 0; i < floor_dx*floor_dy; i++) {
    h->floor_plan[i] = floor_plan[i];
    h->floor_mask[i] = floor_mask[i];
    h->ceiling[i] = ceiling[i];
    h->floor[i] = floor[i];
    h->staircase_base[i] = staircase_base[i];
  }

  for (i = 0; i < floor_dx*floor_dy*2; i++) {
    h->walls_hor[i] = walls_hor[i];
    h->walls_ver[i] = walls_ver[i];
  }

  for (i = 0; i < 256; i++)
    h->area[i] = area[i];

  for (i = 0; i < 256; i++)
    h->room_neighbours[i] = room_neighbours[i];

  h->next = NULL;
  if (floors == NULL)
    floors = h;
  else {
    p = floors;
    while (p->next != NULL)
      p = p->next;
    p->next = h;
  }

  return SUCCEEDED;
}


int houseb_create_floor(void) {

  int i;


  floor_plan = (unsigned char *)malloc(floor_dx * floor_dy);
  floor_mask = (unsigned char *)malloc(floor_dx * floor_dy);
  ceiling = (unsigned char *)malloc(floor_dx * floor_dy);
  floor = (unsigned char *)malloc(floor_dx * floor_dy);
  staircase_base = (unsigned char *)malloc(floor_dx * floor_dy);
  walls_hor = (unsigned char *)malloc(floor_dx * floor_dy * 2);
  walls_ver = (unsigned char *)malloc(floor_dx * floor_dy * 2);
  if (floor_plan == NULL || walls_hor == NULL || walls_ver == NULL || floor_mask == NULL) {
    fprintf(stderr, "HOUSEB_CREATE_FLOOR: Out of memory error.\n");
    return FAILED;
  }

  /* clear the plan */
  for (i = 0; i < floor_dx*floor_dy; i++)
    floor_plan[i] = '.';

  /* clear walls, ceiling and floor etc */
  for (i = 0; i < floor_dx*floor_dy*2; i++) {
    walls_hor[i] = ' ';
    walls_ver[i] = ' ';
    ceiling[i] = ' ';
    floor[i] = ' ';
    staircase_base[i] = ' ';
  }

  /* clear neighbours */
  for (i = 0; i < 256; i++)
    room_neighbours[i] = NULL;

  /* init the floor plan mask (free everywhere) */
  for (i = 0; i < floor_dx*floor_dy; i++)
    floor_mask[i] = 1;

  return SUCCEEDED;
}


void houseb_print_floor_plan(void) {

  int x, y, i;


  printf("FLOOR PLAN\n");

  for (i = 0, y = 0; y < floor_dy; y++) {
    for (x = 0; x < floor_dx; x++, i++) {
      printf("%c", floor_plan[i]);
    }
    printf("\n");
  }
}


void houseb_print_ceiling(void) {

  int x, y, i;


  printf("CEILING\n");

  for (i = 0, y = 0; y < floor_dy; y++) {
    for (x = 0; x < floor_dx; x++, i++) {
      printf("%c", ceiling[i]);
    }
    printf("\n");
  }
}


void houseb_print_staircase_base(void) {

  int x, y, i;


  printf("STAIRCASE BASE\n");

  for (i = 0, y = 0; y < floor_dy; y++) {
    for (x = 0; x < floor_dx; x++, i++) {
      printf("%c", staircase_base[i]);
    }
    printf("\n");
  }
}


void houseb_print_floor(void) {

  int x, y, i;


  printf("FLOOR\n");

  for (i = 0, y = 0; y < floor_dy; y++) {
    for (x = 0; x < floor_dx; x++, i++) {
      printf("%c", floor[i]);
    }
    printf("\n");
  }
}


void houseb_print_plan(unsigned char *p) {

  int x, y;


  for (y = 0; y < floor_dy; y++) {
    for (x = 0; x < floor_dx; x++) {
      printf("%c", *(p++));
    }
    printf("\n");
  }
}


void houseb_print_walls(void) {

  int x, y;


  printf("WALLS\n");

  /* horizontal divider walls */
  /*
  for (i = 0, y = 0; y < floor_dy; y++) {
    for (x = 0; x < floor_dx*2; x++, i++) {
      printf("%c", walls_hor[i]);
    }
    printf("\n");
  }
  */
  /* vertical divider walls */
  /*
  for (i = 0, y = 0; y < floor_dy*2; y++) {
    for (x = 0; x < floor_dx; x++, i++) {
      printf("%c", walls_ver[i]);
    }
    printf("\n");
  }
  */

  /* both combined */
  for (y = 0; y < floor_dy*2; y++) {
    for (x = 0; x < floor_dx*2; x++) {
      if (walls_hor[(y>>1)*floor_dx*2 + x] == 'd')
	printf("d");
      else if (walls_hor[(y>>1)*floor_dx*2 + x] == 'o')
	printf("o");
      else if (walls_ver[y*floor_dx + (x>>1)] == 'D')
	printf("D");
      else if (walls_ver[y*floor_dx + (x>>1)] == 'O')
	printf("O");
      else if (walls_hor[(y>>1)*floor_dx*2 + x] == ' ')
	printf("%c", walls_ver[y*floor_dx + (x>>1)]);
      else
	printf("%c", walls_hor[(y>>1)*floor_dx*2 + x]);
    }
    printf("\n");
  }
}


int houseb_add_room(void) {

  int x, y;


  for (y = 0; y < floor_dy; y++) {
    for (x = 0; x < floor_dx; x++) {
      if (floor_plan[y*floor_dx + x] == '.') {
	if (houseb_insert_room(x, y) == SUCCEEDED)
	  return SUCCEEDED;
      }
    }
  }
  return FAILED;
}


int houseb_insert_room(int x, int y) {

  int dx, dy, a, b;


  dx = floor_dx;
  dy = floor_dy;
  for (a = y; a < floor_dy; a++) {
    for (b = x; b < floor_dx; b++) {
      if (floor_plan[a*floor_dx + b] != '.') {
	if (b < dx) {
	  dx = b;
	  b = floor_dx;
	}
      }
    }
  }
  dx -= x;
  dy -= y;

  if (dx < ROOM_MIN_DX || dy < ROOM_MIN_DY)
    return FAILED;

  if (dx > ROOM_MAX_DX)
    dx = ROOM_MAX_DX;
  if (dy > ROOM_MAX_DX)
    dy = ROOM_MAX_DX;

  a = (rand() % (dx-ROOM_MIN_DX+1)) + ROOM_MIN_DX;
  b = (rand() % (dy-ROOM_MIN_DY+1)) + ROOM_MIN_DY;

  for (dy = y; dy < b+y; dy++) {
    for (dx = x; dx < a+x; dx++) {
      floor_plan[dy*floor_dx + dx] = room;
    }
  }

  room++;

  /* a...z all used up? -> A */
  if (room == 'z'+1)
    room = 'A';
  /* A...Z all used up? -> 0 */
  if (room == 'Z'+1)
    room = '0';

  return SUCCEEDED;
}


void houseb_fill_holes(void) {

  int x, y;


  for (y = 0; y < floor_dy; y++) {
    for (x = 0; x < floor_dx; x++) {
      /* a loose block? */
      if (floor_plan[y*floor_dx + x] == '.') {
	if (y > 0)
	  floor_plan[y*floor_dx + x] = floor_plan[(y-1)*floor_dx + x];
	else if (x > 0)
	  floor_plan[y*floor_dx + x] = floor_plan[y*floor_dx + x-1];
      }
    }
  }
}


void houseb_compute_room_areas(void) {

  int i;


  /* clear room area table */
  for (i = 0; i < 256; i++)
    area[i] = 0;

  /* compute the areas */
  for (i = 0; i < floor_dy*floor_dx; i++)
    area[floor_plan[i]]++;

  area[(unsigned int)'.'] = 0;
}


void houseb_fix_anomalities(void) {

  int i;


  for (i = 0; i < floor_dy; i++) {
    if (floor_plan[floor_dx*i + floor_dx-1] != floor_plan[floor_dx*i + floor_dx-2])
      floor_plan[floor_dx*i + floor_dx-1] = floor_plan[floor_dx*i + floor_dx-2];
  }
}


void houseb_create_walls(void) {

  unsigned char c;
  int x, y;


  /* first get walls dividing rooms horizontally ( | ) */
  for (y = 0; y < floor_dy; y++) {
    for (x = 0; x < floor_dx; x++) {
      c = floor_plan[y*floor_dx + x];
      /* check the left side of the block */
      if ((x == 0 && c != '.') ||
	  (x > 0 && c != '.' && c != floor_plan[y*floor_dx + x - 1])) {
	walls_hor[y*floor_dx*2 + x*2] = 'l';
      }
      /* check the right side of the block */
      if ((x == (floor_dx - 1) && c != '.') ||
	  (c != '.' && c != floor_plan[y*floor_dx + x + 1])) {
	walls_hor[y*floor_dx*2 + x*2 + 1] = 'r';
      }
    }
  }

  /* next get walls dividing rooms vertically ( - ) */
  for (x = 0; x < floor_dx; x++) {
    for (y = 0; y < floor_dy; y++) {
      c = floor_plan[y*floor_dx + x];
      /* check the upper side of the block */
      if ((y == 0 && c != '.') ||
	  (y > 0 && c != '.' && c != floor_plan[(y-1)*floor_dx + x])) {
	walls_ver[y*floor_dx*2 + x] = '-';
      }
      /* check the down side of the block */
      if ((y == (floor_dy - 1) && c != '.') ||
	  (c != '.' && c != floor_plan[(y+1)*floor_dx + x])) {
	walls_ver[y*floor_dx*2 + floor_dx + x] = '_';
      }
    }
  }
}


void houseb_print_neighbours(void) {

  struct neighbour *n;
  unsigned char *c;
  int i, b;


  c = malloc(floor_dx*floor_dy);
  if (c == NULL) {
    fprintf(stderr, "HOUSEB_PRINT_NEIGHBOURS: Out of memory error.\n");
    return;
  }

  printf("NEIGHBOURS\n");

  for (i = 0; i < floor_dx*floor_dy; i++)
    c[i] = '.';

  for (i = 0; i < 256; i++) {
    n = room_neighbours[i];
    while (n != NULL) {
      c[n->y*floor_dx + n->x] = n->room;
      n = n->next;
    }
  }

  for (i = 0; i < floor_dy; i++) {
    for (b = 0; b < floor_dx; b++) {
      printf("%c", c[i*floor_dx + b]);
    }
    printf("\n");
  }

  free(c);
}


int houseb_compute_neighbours(void) {

  struct neighbour *n;
  unsigned char c;
  int x, y;


  for (y = 0; y < floor_dy; y++) {
    for (x = 0; x < floor_dx; x++) {
      c = floor_plan[y*floor_dx + x];
      if (c == '.')
	continue;

      /* check up */
      if (y - 1 >= 0 && floor_plan[(y-1)*floor_dx + x] != c && floor_plan[(y-1)*floor_dx + x] != '.') {
	n = malloc(sizeof(struct neighbour));
	if (n == NULL) {
	  fprintf(stderr, "HOUSEB_COMPUTE_NEIGHBOURS: Out of memory error.\n");
	  return FAILED;
	}
	n->x = x;
	n->y = y;
	n->doored = NO;
	n->neighbour = NEIGHBOUR_UP;
	n->room = floor_plan[(y-1)*floor_dx + x];
	n->next = room_neighbours[(unsigned int)c];
	room_neighbours[(unsigned int)c] = n;
      }

      /* check left */
      if (x - 1 >= 0 && floor_plan[y*floor_dx + (x-1)] != c && floor_plan[y*floor_dx + (x-1)] != '.') {
	n = malloc(sizeof(struct neighbour));
	if (n == NULL) {
	  fprintf(stderr, "HOUSEB_COMPUTE_NEIGHBOURS: Out of memory error.\n");
	  return FAILED;
	}
	n->x = x;
	n->y = y;
	n->doored = NO;
	n->neighbour = NEIGHBOUR_LEFT;
	n->room = floor_plan[y*floor_dx + (x-1)];
	n->next = room_neighbours[(unsigned int)c];
	room_neighbours[(unsigned int)c] = n;
      }

      /* check down */
      if (y + 1 < floor_dy && floor_plan[(y+1)*floor_dx + x] != c && floor_plan[(y+1)*floor_dx + x] != '.') {
	n = malloc(sizeof(struct neighbour));
	if (n == NULL) {
	  fprintf(stderr, "HOUSEB_COMPUTE_NEIGHBOURS: Out of memory error.\n");
	  return FAILED;
	}
	n->x = x;
	n->y = y;
	n->doored = NO;
	n->neighbour = NEIGHBOUR_DOWN;
	n->room = floor_plan[(y+1)*floor_dx + x];
	n->next = room_neighbours[(unsigned int)c];
	room_neighbours[(unsigned int)c] = n;
      }

      /* check right */
      if (x + 1 < floor_dx && floor_plan[y*floor_dx + (x+1)] != c && floor_plan[y*floor_dx + (x+1)] != '.') {
	n = malloc(sizeof(struct neighbour));
	if (n == NULL) {
	  fprintf(stderr, "HOUSEB_COMPUTE_NEIGHBOURS: Out of memory error.\n");
	  return FAILED;
	}
	n->x = x;
	n->y = y;
	n->doored = NO;
	n->neighbour = NEIGHBOUR_RIGHT;
	n->room = floor_plan[y*floor_dx + (x+1)];
	n->next = room_neighbours[(unsigned int)c];
	room_neighbours[(unsigned int)c] = n;
      }
    }
  }

  return SUCCEEDED;
}


void houseb_add_doors(void) {

  unsigned char t[256], q[256], s[256];
  struct neighbour *e;
  int i, x, n, p, y, d, h, c;


  /* go through all the clusters */
  for (c = 0; c < clusters; c++) {

    /* find the starting room (x) */
    for (n = 0, i = 0; i < 256; i++) {
      if (cluster[i] == (c+1) && area[i] > 0) {
	t[i] = 1;
	n++;
      }
      else t[i] = 0;
    }

    if (n == 0)
      continue;

    i = (rand() % n) + 1;

    x = 0;
    while (i > 0) {
      if (cluster[x] == (c+1) && area[x] > 0) {
	i--;
	if (i == 0)
	  break;
      }
      x++;
    }

    /* add one item to the queue (the root) */
    i = 1;
    q[0] = x;
    t[x] = 0;

    while (n > 0) {
      i--;
      x = q[i];

      /* collect the neighbours */
      e = room_neighbours[x];
      p = 0;
      while (e != NULL) {
	/* already listed? */
	for (y = 0; y < p; y++)
	  if (s[y] == e->room)
	    break;
	/* a new entry? */
	if (y == p && t[(unsigned int)e->room] == 1) {
	  t[(unsigned int)e->room] = 0;
	  s[p] = e->room;
	  p++;
	}
	e = e->next;
      }

      /* this room has p neighbouring rooms that are not yet connected to the maze */
      if (p == 0) {
	n--;
	continue;
      }
      y = p;

      while (y > 0) {
	d = rand() % p;
	if (s[d] != 0) {
	  q[i] = s[d];
	  i++;
	  y--;

	  /* count the possible places for a door */
	  e = room_neighbours[x];
	  h = 0;
	  while (e != NULL) {
	    if (e->room == s[d] && e->doored == NO)
	      h++;
	    e = e->next;
	  }

	  /* choose one of them */
	  h = (rand() % h) + 1;

	  e = room_neighbours[x];
	  while (h > 0) {
	    if (e->room == s[d] && e->doored == NO)
	      h--;
	    if (h > 0)
	      e = e->next;
	  }

	  _houseb_add_door(e);
	  s[d] = 0;
	}
      }

      n--;
    }
  }

  /* add a few redundant doors */
  h = 0;
  i = 0;
  for (x = 0; x < 256; x++) {
    e = room_neighbours[x];
    while (e != NULL) {
      if (e->doored == YES)
	h++;
      else
	i++;
      e = e->next;
    }
  }

  /* max 25% more */
  if ((h>>2) == 0)
    return;

  h = rand() % (h>>2);
  h += rand() & 1;

  if (h == 0)
    return;

  printf("%d additional door(s).\n", h);

  while (h > 0) {
    x = (rand() % i) + 1;
    for (p = 0; x > 0; p++) {
      e = room_neighbours[p];
      while (e != NULL) {
	if (e->doored == NO)
	  x--;
	if (x > 0)
	  e = e->next;
	else {
	  if (_houseb_add_door(e) == FAILED) {
	    /* try again! */
	    h++;
	  }
	  e = NULL;
	}
      }
    }
    h--;
    i--;
  }
}


int _houseb_add_door(struct neighbour *n) {

  if (n->doored == YES) {
    fprintf(stderr, "_HOUSEB_ADD_DOOR: Door (%d, %d) exists already!\n", n->x, n->y);
    return FAILED;
  }

  n->doored = YES;

  if (n->neighbour == NEIGHBOUR_UP) {
    if (walls_ver[n->y*floor_dx*2 + n->x] != '-')
      return FAILED;
    walls_ver[n->y*floor_dx*2 + n->x] = 'D';
    walls_ver[n->y*floor_dx*2 - floor_dx + n->x] = 'O';
  }
  else if (n->neighbour == NEIGHBOUR_DOWN) {
    if (walls_ver[n->y*floor_dx*2 + floor_dx + n->x] != '_')
      return FAILED;
    walls_ver[n->y*floor_dx*2 + floor_dx + n->x] = 'O';
    walls_ver[n->y*floor_dx*2 + floor_dx*2 + n->x] = 'D';
  }
  else if (n->neighbour == NEIGHBOUR_LEFT) {
    if (walls_hor[n->y*floor_dx*2 + n->x*2] != 'l')
      return FAILED;
    walls_hor[n->y*floor_dx*2 + n->x*2] = 'o';
    walls_hor[n->y*floor_dx*2 + n->x*2 - 1] = 'd';
  }
  else if (n->neighbour == NEIGHBOUR_RIGHT) {
    if (walls_hor[n->y*floor_dx*2 + n->x*2 + 1] != 'r')
      return FAILED;
    walls_hor[n->y*floor_dx*2 + n->x*2 + 1] = 'd';
    walls_hor[n->y*floor_dx*2 + n->x*2 + 2] = 'o';
  }

  return SUCCEEDED;
}


void houseb_apply_mask(void) {

  int i;


  for (i = 0; i < floor_dx*floor_dy; i++) {
    if (floor_mask[i] != 1)
      floor_plan[i] = '.';
  }
}


void houseb_fix_labels(void) {

  unsigned char n;
  int x, y;


  memset(tmp, 0, floor_dx*floor_dy);

  /* fix the labels (mask might have broken some rooms into two or more rooms) */
  n = 0;
  for (y = 0; y < floor_dy; y++) {
    for (x = 0; x < floor_dx; x++) {
      if (tmp[y*floor_dx + x] == 0 && floor_plan[y*floor_dx + x] != '.') {
	_houseb_fix_label(x, y, n, floor_plan[y*floor_dx + x], tmp);
	n++;
      }
    }
  }

  /* return the labels to alphabets */
  for (y = 0; y < floor_dy; y++) {
    for (x = 0; x < floor_dx; x++) {
      if (floor_plan[y*floor_dx + x] != '.')
	floor_plan[y*floor_dx + x] += 'a';
    }
  }
}


void _houseb_fix_label(int x, int y, unsigned char n, unsigned char o, unsigned char *t) {

  if (x < 0 || x >= floor_dx || y < 0 || y >= floor_dy)
    return;

  if (floor_plan[y*floor_dx + x] == o) {
    t[y*floor_dx + x] = 1;
    floor_plan[y*floor_dx + x] = n;
    /* fix all the hour neighbours as well, if they belong to this room */
    _houseb_fix_label(x + 1, y, n, o, t);
    _houseb_fix_label(x - 1, y, n, o, t);
    _houseb_fix_label(x, y + 1, n, o, t);
    _houseb_fix_label(x, y - 1, n, o, t);
  }
}


void houseb_fix_clusters(void) {

  unsigned char n;
  int x, y, z, a[256], i, c;


  memset(tmp, 0, floor_dx*floor_dy);

  for (i = 0; i < 256; i++)
    cluster[i] = 0;

  /* compute room clusters (four-neighbour rule) */
  n = 1;
  for (y = 0; y < floor_dy; y++) {
    for (x = 0; x < floor_dx; x++) {
      if (tmp[y*floor_dx + x] == 0 && floor_plan[y*floor_dx + x] != '.') {
	/* create a new cluster */
	_houseb_fix_cluster(x, y, tmp, n);
	n++;
      }
    }
  }

  for (x = 0; x < 256; x++)
    a[x] = 0;

  /* compute cluster areas */
  for (x = 0; x < floor_dx*floor_dy; x++)
    a[tmp[x]]++;

  /* compute clusters */
  for (x = 1, clusters = 0; x < 256; x++)
    if (a[x] != 0)
      clusters++;

  printf("%d cluster(s) -> ", clusters);

  /* drop very small clusters */
  for (i = 0, x = 1; x < 256; x++) {
    if (a[x] == 0)
      continue;
    else if (a[x] < 4) {
      a[x] = 0;
      clusters--;
      i++;
      for (y = 0; y < floor_dx*floor_dy; y++)
	if (tmp[y] == x)
	  floor_plan[y] = '.';
    }
  }

  printf("dropped %d cluster(s) (too small).\n", i);

  /* assign cluster ids to rooms */
  for (y = 0; y < floor_dx*floor_dy; y++) {
    if (floor_plan[y] != '.') {
      cluster[floor_plan[y]] = tmp[y];
    }
  }

  /* fix the cluster numbers */
  for (y = 0, x = cluster[0]; y < 256; y++) {
    if (cluster[y] > x)
      x = cluster[y];
  }

  for (y = 1, i = 1; y <= x; y++) {
    for (c = 0, z = 0; z < 256; z++) {
      if (cluster[z] == y) {
	cluster[z] = i;
	c = 1;
      }
    }
    if (c == 1)
      i++;
  }
}


void _houseb_fix_cluster(int x, int y, unsigned char *t, unsigned char n) {

  if (x < 0 || x >= floor_dx || y < 0 || y >= floor_dy)
    return;

  if (t[y*floor_dx + x] != 0 || floor_plan[y*floor_dx + x] == '.')
    return;

  t[y*floor_dx + x] = n;

  _houseb_fix_cluster(x + 1, y, t, n);
  _houseb_fix_cluster(x - 1, y, t, n);
  _houseb_fix_cluster(x, y + 1, t, n);
  _houseb_fix_cluster(x, y - 1, t, n);
}


void houseb_drop_random_rooms(void) {

  int i, n, x, y, d;


  /* drop random amount of rooms */
  i = rand() % 10;
  if (i > 2 || i == 0)
    return;

  d = i;
  houseb_compute_room_areas();
  houseb_compute_rooms();

  /* don't drop them all! */
  if (rooms - i <= 0)
    return;

  /* drop the rooms */
  while (i > 0) {
    n = (rand() % rooms) + 1;
    for (x = 0; x < 256; x++) {
      if (area[x] > 0)
	n--;
      if (n == 0) {
	for (y = 0; y < floor_dy*floor_dx; y++) {
	  if (floor_plan[y] == x)
	    floor_plan[y] = '.';
	}
	x = 256;
	i--;
      }
    }
  }

  printf("%d room(s) were dropped.\n", d);
}


void houseb_compute_rooms(void) {

  int i;


  rooms = 0;
  for (i = 0; i < 256; i++)
    if (area[i] > 0)
      rooms++;
}


void houseb_compute_ceiling(void) {

  int i;


  for (i = 0; i < floor_dy*floor_dx; i++)
    if (floor_plan[i] != '.')
      ceiling[i] = '*';
}


void houseb_compute_floor(void) {

  int i;


  for (i = 0; i < floor_dy*floor_dx; i++)
    if (floor_plan[i] != '.')
      floor[i] = '*';
}


void houseb_compute_windowed_walls(void) {

  unsigned char *c;
  int x, y;


  /* handle 'l' and 'r' */
  c = walls_hor;
  for (y = 0; y < floor_dy; y++) {
    for (x = 0; x < floor_dx*2; x++) {
      /* left wall can be windowed? */
      if (*c == 'l') {
	if (x == 0 || *(c-1) != 'r')
	  *c = 'L';
      }
      /* right wall can be windowed? */
      if (*c == 'r') {
	if (x == floor_dx*2-1 || *(c+1) != 'l')
	  *c = 'R';
      }
      c++;
    }
  }

  /* handle '-' and '_' */
  c = walls_ver;
  for (y = 0; y < floor_dy*2; y++) {
    for (x = 0; x < floor_dx; x++) {
      /* upper wall can be windowed? */
      if (*c == '-') {
	if (y == 0 || *(c-floor_dx) != '_')
	  *c = 'Y';
      }
      /* lower wall can be windowed? */
      if (*c == '_') {
	if (y == floor_dy*2-1 || *(c+floor_dx) != '-')
	  *c = 'A';
      }
      c++;
    }
  }
}


void houseb_compute_staircase_base(void) {

  int y, x, i, d;


  for (i = 0, y = 0; y < floor_dy; y++) {
    for (x = 0; x < floor_dx; x++, i++) {
      if (floor_plan[i] != '.') {
	d = 0;
	if (x > 0 && floor_plan[i-1] != floor_plan[i])
	  d++;
	else if (x == 0)
	  d++;
	if (x < floor_dx-1 && floor_plan[i+1] != floor_plan[i])
	  d++;
	else if (x == floor_dx-1)
	  d++;
	if (y > 0 && floor_plan[i-floor_dx] != floor_plan[i])
	  d++;
	else if (y == 0)
	  d++;
	if (y < floor_dy-1 && floor_plan[i+floor_dx] != floor_plan[i])
	  d++;
	else if (y == floor_dy-1)
	  d++;

	if (d > 0)
	  staircase_base[i] = '*';
	else
	  staircase_base[i] = '.';
      }
      else
	staircase_base[i] = '.';
    }
  }
}


/* check that there are no doors around this block */
int _houseb_no_doors(int x, int y, struct house_floor *h) {

  if (h->walls_hor[y*floor_dx*2 + x*2 + 0] == 'o')
    return FAILED;
  if (h->walls_hor[y*floor_dx*2 + x*2 + 1] == 'd')
    return FAILED;
  if (h->walls_ver[y*floor_dx*2 + x] == 'D')
    return FAILED;
  if (h->walls_ver[y*floor_dx*2 + x + floor_dx] == 'O')
    return FAILED;

  return SUCCEEDED;
}


int houseb_place_staircases(void) {

  unsigned char *c, *d;
  struct house_floor *a, *b;
  int i, x, y;


  c = malloc(floor_dx*floor_dy);
  d = malloc(floor_dx*floor_dy);
  if (c == NULL || d == NULL) {
    fprintf(stderr, "HOUSEB_PLACE_STAIRCASES: Out of memory error.\n");
    return FAILED;
  }

  if (floors == NULL)
    return FAILED;

  a = floors;
  b = floors->next;
  while (b != NULL) {
    /* find suitable places for staircases (vertical, ^up) */
    for (i = 0, y = 0; y < floor_dy; y++) {
      for (x = 0; x < floor_dx; x++) {
	if (a->staircase_base[i] == '*' && y > 1 && y < floor_dy-1 && a->staircase_base[i-floor_dx] == '*' && a->staircase_base[i+floor_dx] == '*' &&
	    b->floor_plan[i] != '.' && b->floor_plan[i] == b->floor_plan[i-floor_dx] &&
	    b->floor_plan[i] == b->floor_plan[i-(2*floor_dx)] &&
	    a->floor_plan[i] != '.' && a->floor_plan[i] == a->floor_plan[i-floor_dx] && a->floor_plan[i+floor_dx] == a->floor_plan[i] &&
	    _houseb_no_doors(x, y, b) == SUCCEEDED && _houseb_no_doors(x, y-1, b) == SUCCEEDED &&
	    _houseb_no_doors(x, y, a) == SUCCEEDED && _houseb_no_doors(x, y-1, a) == SUCCEEDED)
	  c[i] = '*';
	else
	  c[i] = '.';
	i++;
      }
    }

    /* find suitable places for staircases (horizontal, ^left) */
    for (i = 0, y = 0; y < floor_dy; y++) {
      for (x = 0; x < floor_dx; x++) {
	if (a->staircase_base[i] == '*' && x > 1 && x < floor_dx-1 && a->staircase_base[i-1] == '*' && a->staircase_base[i+1] == '*' &&
	    b->floor_plan[i] != '.' && b->floor_plan[i] == b->floor_plan[i-1] &&
	    b->floor_plan[i] == b->floor_plan[i-2] &&
	    a->floor_plan[i] != '.' && a->floor_plan[i] == a->floor_plan[i-1] && a->floor_plan[i] == a->floor_plan[i+1] &&
	    _houseb_no_doors(x, y, b) == SUCCEEDED && _houseb_no_doors(x-1, y, b) == SUCCEEDED &&
	    _houseb_no_doors(x, y, a) == SUCCEEDED && _houseb_no_doors(x-1, y, a) == SUCCEEDED)
	  d[i] = '*';
	else
	  d[i] = '.';
	i++;
      }
    }

    /* count the suitable places */
    i = 0;
    for (x = 0; x < floor_dx*floor_dy; x++) {
      if (c[x] == '*')
	i++;
      if (d[x] == '*')
	i++;
    }

    if (i == 0) {
      fprintf(stderr, "HOUSEB_PLACE_STAIRCASES: No suitable place for stairs!\n");
      return FAILED;
    }

    /* take a random place */
    i = (rand() % i) + 1;

    for (x = 0; x < floor_dx*floor_dy; x++) {
      if (c[x] == '*')
	i--;
      if (i == 0)
	break;
    }

    /* it's a vertical one */
    if (i == 0) {
      a->ceiling[x] = 'H';
      a->ceiling[x-floor_dx] = 'H';
      b->floor[x] = 'H';
      b->floor[x-floor_dx] = 'H';
      b->staircase_base[x] = '.';
      b->staircase_base[x-floor_dx] = '.';
      a->staircase_base[x] = 'S';
    }
    /* it's a horizontal one */
    else {
      for (x = 0; x < floor_dx*floor_dy; x++) {
	if (d[x] == '*')
	  i--;
	if (i == 0)
	  break;
      }
      a->ceiling[x] = 'H';
      a->ceiling[x-1] = 'H';
      b->floor[x] = 'H';
      b->floor[x-1] = 'H';
      b->staircase_base[x] = '.';
      b->staircase_base[x-1] = '.';
      a->staircase_base[x] = 'L';
    }

    printf("STAIRCASE BASE\n");
    houseb_print_plan(a->staircase_base);

    a = b;
    b = b->next;
  }

  free(c);
  free(d);

  return SUCCEEDED;
}


void houseb_output_house(void) {

  struct house_floor *fl;
  unsigned char *c;
  char name[] = "out.txt";
  int x, y, z;
  FILE *f;


  f = fopen(name, "wb");
  if (f == NULL) {
    fprintf(stderr, "HOUSEB_OUTPUT_HOUSE: Could not open \"%s\".\n", name);
    return;
  }

  /* output floor by floor */
  fl = floors;
  z = 0;
  while (fl != NULL) {
    /* output horizontal walls */
    c = fl->walls_ver;
    for (y = 0; y < floor_dy; y++) {
      for (x = 0; x < floor_dx; x++) {
	if (*c == '-')
	  fprintf(f, "landscape %s/wall_u.lan %f %f %f\n", landir, x*BLOCK_DX + house_x, z*BLOCK_DZ + house_z, y*BLOCK_DY + house_y);
	if (*c == 'Y')
	  fprintf(f, "landscape %s/wall_u_win.lan %f %f %f\n", landir, x*BLOCK_DX + house_x, z*BLOCK_DZ + house_z, y*BLOCK_DY + house_y);
	c++;
      }
      for (x = 0; x < floor_dx; x++) {
	if (*c == '_')
	  fprintf(f, "landscape %s/wall_d.lan %f %f %f\n", landir, x*BLOCK_DX + house_x, z*BLOCK_DZ + house_z, y*BLOCK_DY + house_y);
	if (*c == 'A')
	  fprintf(f, "landscape %s/wall_d_win.lan %f %f %f\n", landir, x*BLOCK_DX + house_x, z*BLOCK_DZ + house_z, y*BLOCK_DY + house_y);
	c++;
      }
    }

    /* output vertical walls */
    c = fl->walls_hor;
    for (y = 0; y < floor_dy; y++) {
      for (x = 0; x < floor_dx; x++) {
	if (*c == 'l')
	  fprintf(f, "landscape %s/wall_l.lan %f %f %f\n", landir, x*BLOCK_DX + house_x, z*BLOCK_DZ + house_z, y*BLOCK_DY + house_y);
	if (*c == 'L')
	  fprintf(f, "landscape %s/wall_l_win.lan %f %f %f\n", landir, x*BLOCK_DX + house_x, z*BLOCK_DZ + house_z, y*BLOCK_DY + house_y);
	c++;
	if (*c == 'r')
	  fprintf(f, "landscape %s/wall_r.lan %f %f %f\n", landir, x*BLOCK_DX + house_x, z*BLOCK_DZ + house_z, y*BLOCK_DY + house_y);
	if (*c == 'R')
	  fprintf(f, "landscape %s/wall_r_win.lan %f %f %f\n", landir, x*BLOCK_DX + house_x, z*BLOCK_DZ + house_z, y*BLOCK_DY + house_y);
	c++;
      }
    }

    /* output ceiling blocks */
    c = fl->ceiling;
    for (y = 0; y < floor_dy; y++) {
      for (x = 0; x < floor_dx; x++) {
	if (y > 0 && y < floor_dy-1 && x > 0 && x < floor_dx-1 && *(c+1) == '*' && *(c-1) == '*' &&
	    *(c-floor_dx) == '*' && *(c+floor_dx) == '*' && *c == '*')
	  fprintf(f, "landscape %s/ceiling_optimized.lan %f %f %f\n", landir, x*BLOCK_DX + house_x, z*BLOCK_DZ + house_z, y*BLOCK_DY + house_y);
	else if (*c == '*')
	  fprintf(f, "landscape %s/ceiling.lan %f %f %f\n", landir, x*BLOCK_DX + house_x, z*BLOCK_DZ + house_z, y*BLOCK_DY + house_y);
	c++;
      }
    }

    /* output floor blocks */
    c = fl->floor;
    for (y = 0; y < floor_dy; y++) {
      for (x = 0; x < floor_dx; x++) {
	if (y > 0 && y < floor_dy-1 && x > 0 && x < floor_dx-1 && *(c+1) == '*' && *(c-1) == '*' &&
	    *(c-floor_dx) == '*' && *(c+floor_dx) == '*' && *c == '*')
	  fprintf(f, "landscape %s/floor_optimized.lan %f %f %f\n", landir, x*BLOCK_DX + house_x, z*BLOCK_DZ + house_z, y*BLOCK_DY + house_y);
	if (*c == '*')
	  fprintf(f, "landscape %s/floor.lan %f %f %f\n", landir, x*BLOCK_DX + house_x, z*BLOCK_DZ + house_z, y*BLOCK_DY + house_y);
	c++;
      }
    }

    /* output staircases */
    c = fl->staircase_base;
    for (y = 0; y < floor_dy; y++) {
      for (x = 0; x < floor_dx; x++) {
	if (*c == 'S')
	  fprintf(f, "landscape %s/stairs_u.lan %f %f %f\n", landir, x*BLOCK_DX + house_x, z*BLOCK_DZ + house_z, y*BLOCK_DY + house_y);
	if (*c == 'L')
	  fprintf(f, "landscape %s/stairs_l.lan %f %f %f\n", landir, x*BLOCK_DX + house_x, z*BLOCK_DZ + house_z, y*BLOCK_DY + house_y);
	c++;
      }
    }

    fl = fl->next;
    z++;
  }

  fclose(f);
}
