
/* (c) 2001 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include "defines.h"
#include "oload.h"
#include "txt.h"


struct vrml_object *objects_first = NULL, *objects_last = NULL;
struct vrml_shape *shapes_first = NULL, *shapes_last = NULL, *sha;
double sc_r, sc_g, sc_b, dc_r, dc_g, dc_b;
double *double_tmp, dob, *coordinate_point_list;
double *texture_coordinate_point_list;
char txt[256];
int double_tmp_size = 1024, coordinate_points;
int txt_d = OFF, sc_d = OFF, dc_d = OFF, cop_d = OFF, cip_d = OFF, tcd_d = OFF, tci_d = OFF;
int *int_tmp, *coordindex_point_list, coordindex_points, int_tmp_size = 1024;
int texture_coordinate_points;
int texture_coordindex_points, *texture_coordindex_point_list;
int list_end = OFF;


int object_load(char *name) {

  struct vrml_object *v;
  static int objects = 0;
  FILE *f;
  char *b;
  int s;


  f = fopen(name, "rb");
  if (f == NULL) {
    fprintf(stderr, "OBJECT_LOAD: Could not open file \"%s\".\n", name);
    return FAILED;
  }

  fseek(f, 0, SEEK_END);
  s = ftell(f);
  fseek(f, 0, SEEK_SET);

  b = malloc(s);
  if (b == NULL) {
    fprintf(stderr, "OBJECT_LOAD: Out of memory.\n");
    fclose(f);
    return FAILED;
  }

  fread(b, 1, s, f);
  fclose(f);

  shapes_first = NULL;
  shapes_last = NULL;
  object_parse_vrml(b, s);

  free(b);

  v = malloc(sizeof(struct vrml_object));
  if (v == NULL) {
    fprintf(stderr, "OBJECT_LOAD: Out of memory.\n");
    return FAILED;
  }

  v->next = NULL;
  v->shape = shapes_first;
  v->id = objects;
  objects++;

  if (objects_first == NULL) {
    objects_first = v;
    objects_last = v;
  }
  else {
    objects_last->next = v;
    objects_last = v;
  }

  return SUCCEEDED;
}


int object_parse_vrml(char *b, int s) {

  char r[256];
  int i = 0, y, d, shapes = 0;


  double_tmp = malloc(sizeof(double) * double_tmp_size);
  if (double_tmp == NULL) {
    fprintf(stderr, "OBJECT_PARSE_VRML: Out of memory.\n");
    return FAILED;
  }
  int_tmp = malloc(sizeof(int) * int_tmp_size);
  if (int_tmp == NULL) {
    fprintf(stderr, "OBJECT_PARSE_VRML: Out of memory.\n");
    free(double_tmp);
    return FAILED;
  }

  txt[0] = 0;

  while (i < s) {
    object_get_next_token(b, r, &i, &y, s);

    /* shape? */
    if (strcmp(r, "Shape") == 0) {
      if (shapes != 0) {
	add_new_shape();
      }
      shapes++;
    }
    /* texture? */
    else if (strcmp(r, "ImageTexture") == 0) {
      object_get_next_token(b, r, &i, &y, s);
      object_get_next_token(b, r, &i, &y, s);
      object_get_next_token(b, r, &i, &y, s);
      if (r[0] == '"') {
	strcpy(txt, &r[1]);
	txt[y - 2] = 0;
      }
      else
	strcpy(txt, r);
      if (txt[0] == 'f' && txt[1] == 'i' && txt[2] == 'l' && txt[3] == 'e' && txt[4] == ':') {
	strcpy(r, &txt[7]);
	strcpy(txt, r);
      }
      txt_d = ON;
    }
    /* specular color? */
    else if (strcmp(r, "specularColor") == 0) {
      object_get_next_token(b, r, &i, &y, s);
      obtain_double(r, &sc_r);
      object_get_next_token(b, r, &i, &y, s);
      obtain_double(r, &sc_g);
      object_get_next_token(b, r, &i, &y, s);
      obtain_double(r, &sc_b);
      sc_d = ON;
    }
    /* diffuse color? */
    else if (strcmp(r, "diffuseColor") == 0) {
      object_get_next_token(b, r, &i, &y, s);
      obtain_double(r, &dc_r);
      object_get_next_token(b, r, &i, &y, s);
      obtain_double(r, &dc_g);
      object_get_next_token(b, r, &i, &y, s);
      obtain_double(r, &dc_b);
      dc_d = ON;
    }
    /* coordinate? */
    else if (strcmp(r, "Coordinate") == 0) {
      object_get_next_token(b, r, &i, &y, s);
      object_get_next_token(b, r, &i, &y, s);
      if (strcmp(r, "point") != 0) {
	fprintf(stderr, "OBJECT_PARSE_VRML: Syntax error in \"Coordinate\".\n");
	return FAILED;
      }
      object_get_next_token(b, r, &i, &y, s);
      d = 0;
      while (list_end == OFF) {
	object_get_next_token(b, r, &i, &y, s);
	if (y == 0 && list_end == ON)
	  break;
	obtain_double(r, &dob);
	double_tmp[d] = dob;
	d++;
	object_get_next_token(b, r, &i, &y, s);
	obtain_double(r, &dob);
	double_tmp[d] = dob;
	d++;
	object_get_next_token(b, r, &i, &y, s);
	obtain_double(r, &dob);
	double_tmp[d] = dob;
	d++;
      }
      coordinate_points = d / 3;
      coordinate_point_list = malloc(sizeof(double) * d);
      if (coordinate_point_list == NULL) {
	fprintf(stderr, "OBJECT_PARSE_VRML: Out of memory.\n");
	return FAILED;
      }
      memcpy(coordinate_point_list, double_tmp, sizeof(double) * d);
      cop_d = ON;
    }
    /* coordindex? */
    else if (strcmp(r, "coordIndex") == 0) {
      object_get_next_token(b, r, &i, &y, s);
      d = 0;
      while (list_end == OFF) {
	object_get_next_token(b, r, &i, &y, s);
	if (list_end == ON)
	  break;
	obtain_integer(r, &y);
	if (y == -1)
	  continue;
	int_tmp[d] = y;
	d++;
      }
      coordindex_points = d;
      coordindex_point_list = malloc(sizeof(int) * d);
      if (coordindex_point_list == NULL) {
	fprintf(stderr, "OBJECT_PARSE_VRML: Out of memory.\n");
	return FAILED;
      }
      memcpy(coordindex_point_list, int_tmp, sizeof(int) * d);
      cip_d = ON;
    }
    /* texturecoordinate? */
    else if (strcmp(r, "TextureCoordinate") == 0) {
      object_get_next_token(b, r, &i, &y, s);
      object_get_next_token(b, r, &i, &y, s);
      if (strcmp(r, "point") != 0) {
	fprintf(stderr, "OBJECT_PARSE_VRML: Syntax error in \"TextureCoordinate\".\n");
	return FAILED;
      }
      object_get_next_token(b, r, &i, &y, s);
      d = 0;
      while (list_end == OFF) {
	object_get_next_token(b, r, &i, &y, s);
	if (list_end == ON)
	  break;
	obtain_double(r, &dob);
	double_tmp[d] = dob;
	d++;
      }
      texture_coordinate_points = d / 2;
      texture_coordinate_point_list = malloc(sizeof(double) * d);
      if (texture_coordinate_point_list == NULL) {
	fprintf(stderr, "OBJECT_PARSE_VRML: Out of memory.\n");
	return FAILED;
      }
      memcpy(texture_coordinate_point_list, double_tmp, sizeof(double) * d);
      tcd_d = ON;
    }
    /* texturecoordindex? */
    else if (strcmp(r, "texCoordIndex") == 0) {
      object_get_next_token(b, r, &i, &y, s);
      d = 0;
      while (list_end == OFF) {
	object_get_next_token(b, r, &i, &y, s);
	if (list_end == ON)
	  break;
	obtain_integer(r, &y);
	if (y == -1)
	  continue;
	int_tmp[d] = y;
	d++;
      }
      texture_coordindex_points = d;
      texture_coordindex_point_list = malloc(sizeof(int) * d);
      if (texture_coordindex_point_list == NULL) {
	fprintf(stderr, "OBJECT_PARSE_VRML: Out of memory.\n");
	return FAILED;
      }
      memcpy(texture_coordindex_point_list, int_tmp, sizeof(int) * d);
      tci_d = ON;
    }
  }

  /* final addition */
  add_new_shape();

  free(int_tmp);
  free(double_tmp);

#if DEBUG == 1
  sha = shapes_first;
  d = 0;
  while (sha != NULL) {
    printf("shape %d\n", d);
    printf("texture: %s\n", sha->txt);
    if (sha->sc_d == ON)
      printf("specular color: %f %f %f\n", sha->sc_r, sha->sc_g, sha->sc_b);
    else
      printf("diffuse color: none\n");
    if (sha->dc_d == ON)
      printf("diffuse color: %f %f %f\n", sha->dc_r, sha->dc_g, sha->dc_b);
    else
      printf("diffuse color: none\n");

    if (sha->cop_d == ON) {
      printf("coordinate points:\n");
      y = sha->coordinate_points;
      coordinate_point_list = sha->coordinate_point_list;
      while (y > 0) {
	printf(" %f %f %f,\n", coordinate_point_list[0], coordinate_point_list[1],
	       coordinate_point_list[2]);
	y--;
	coordinate_point_list += 3;
      }
    }
    else
      printf("coordinate points: none\n");

    if (sha->cip_d == ON) {
      printf("coordinate index points:\n");
      y = sha->coordindex_points;
      coordindex_point_list = sha->coordindex_point_list;
      while (y > 0) {
	printf(" %d,\n", coordindex_point_list[0]);
	y--;
	coordindex_point_list++;
      }
    }
    else
      printf("coordindex points: none\n");

    if (sha->tcd_d == ON) {
      printf("texture coordinate points:\n");
      y = sha->texture_coordinate_points;
      texture_coordinate_point_list = sha->texture_coordinate_point_list;
      while (y > 0) {
	printf(" %f %f,\n", texture_coordinate_point_list[0], texture_coordinate_point_list[1]);
	y--;
	texture_coordinate_point_list += 2;
      }
    }
    else
      printf("texture_coordinate points: none\n");

    if (sha->tci_d == ON) {
      printf("texture coordinate index points:\n");
      y = sha->texture_coordindex_points;
      texture_coordindex_point_list = sha->texture_coordindex_point_list;
      while (y > 0) {
	printf(" %d,\n", texture_coordindex_point_list[0]);
	y--;
	texture_coordindex_point_list++;
      }
    }
    else
      printf("texture coordindex points: none\n");

    sha = sha->next;
    d++;
  }
#endif

  return SUCCEEDED;
}


int object_get_next_token(char *b, char *r, int *s, int *k, int m) {

  list_end = OFF;

  while (*s < m && (*(b+*s) == ',' || *(b+*s) == ' ' || *(b+*s) == 0x0a || *(b+*s) == 0x0d))
    (*s)++;

  *k = 0;

  while (*s < m && *(b+*s) != ',' && *(b+*s) != ' ' && *(b+*s) != 0x0a && *(b+*s) != 0x0d) {
    *(r+*k) = *(b+*s);
    (*s)++;
    (*k)++;
  }

  if (*(r+*k-1) == ']') {
    (*k)--;
    list_end = ON;
  }

  *(r+*k) = 0;

  return SUCCEEDED;
}


int obtain_integer(char *b, int *i) {

  int m;


  if (*b == '-') {
    m = -1;
    b++;
  }
  else
    m = 1;

  *i = 0;
  while (*b != 0) {
    *i = *i * 10 + *b - '0';
    b++;
  }

  *i = *i * m;

  return SUCCEEDED;
}


int obtain_double(char *b, double *d) {

  int i, m, c;


  if (*b == '-') {
    m = -1;
    b++;
  }
  else
    m = 1;

  i = 0;
  while (*b != 0 && *b != '.') {
    i = i * 10 + *b - '0';
    b++;
  }

  if (*b == 0) {
    *d = i * m;
    return SUCCEEDED;
  }

  b++;
  *d = 0;
  c = 10;
  while (*b != 0) {
    *d += ((double)(*b - '0')) / c;
    b++;
    c *= 10;
  }

  *d = (i + *d) * m;
  return SUCCEEDED;
}


int add_new_shape(void) {

  sha = malloc(sizeof(struct vrml_shape));
  if (sha == NULL) {
    fprintf(stderr, "ADD_NEW_SHAPE: Out of memory.\n");
    return FAILED;
  }

  if (shapes_first == NULL) {
    shapes_first = sha;
    shapes_last = sha;
  }
  else {
    shapes_last->next = sha;
    shapes_last = sha;
  }

  sha->next = NULL;
  sha->txt_d = txt_d;
  sha->sc_d = sc_d;
  sha->dc_d = dc_d;
  sha->sc_r = sc_r;
  sha->sc_g = sc_g;
  sha->sc_b = sc_b;
  sha->dc_r = dc_r;
  sha->dc_g = dc_g;
  sha->dc_b = dc_b;
  sha->cop_d = cop_d;
  sha->cip_d = cip_d;
  sha->tcd_d = tcd_d;
  sha->tci_d = tci_d;
  sha->coordinate_points = coordinate_points;
  sha->coordinate_point_list = coordinate_point_list;
  sha->coordindex_points = coordindex_points;
  sha->coordindex_point_list = coordindex_point_list;
  sha->texture_coordinate_points = texture_coordinate_points;
  sha->texture_coordinate_point_list = texture_coordinate_point_list;
  sha->texture_coordindex_points = texture_coordindex_points;
  sha->texture_coordindex_point_list = texture_coordindex_point_list;

  if (txt_d == ON) {
    if (txt_load(txt, &(sha->txt_id)) == FAILED)
      return FAILED;
  }

  txt_d = OFF;
  sc_d = OFF;
  dc_d = OFF;
  cop_d = OFF;
  cip_d = OFF;
  tcd_d = OFF;
  tci_d = OFF;

  return SUCCEEDED;
}
