
/* (c) 2003 ville helin <vhelin@cc.hut.fi> */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <SDL/SDL_video.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "defines.h"
#include "../common/defines.h"
#include "../common/mesh2.h"
#include "../common/quaternion.h"
#include "txt.h"
#include "gui.h"
#include "main.h"


extern int status, view, view_mode;
extern double player_x, player_y;
struct mesh2 **meshes = NULL;
double distance = 50;
int nmeshes;


int main(int argc, char *argv[]) {

  int i;


  if (argc < 2) {
    printf("usage: %s <me2 files...>\n", argv[0]);
    return 1;
  }

  nmeshes = argc - 1;

  init_window_main();

  txt_init("../data/txt/");

  meshes = malloc(sizeof(struct mesh2 *) * nmeshes);
  if (meshes == NULL) {
    fprintf(stderr, "MAIN: Out of memory error.\n");
    return 1;
  }

  /* load the objects */
  for (i = 1; i < argc; i++) {
    meshes[i-1] = NULL;
    if (mesh2_read_binary(argv[i], &meshes[i-1], YES) == FAILED)
      return 1;
  }

  glViewport(0, 0, SCREEN_X, SCREEN_Y);

  SDL_ShowCursor(SDL_DISABLE);
  SDL_SetEventFilter(sdl_filter_events);

  while (1) {
    keyboard_io();
    ogl_draw();
    if (status & STATUS_QUIT)
      break;
    if (status & STATUS_DISTANCE_MINUS) {
      if (distance > 4)
	distance -= 0.1;
    }
    if (status & STATUS_DISTANCE_PLUS) {
      if (distance < 100)
	distance += 0.1;
    }
  }

  SDL_ShowCursor(SDL_ENABLE);
  SDL_Quit();

  return 0;
}


int ogl_draw(void) {

  GLfloat rm[16];
  GLint vp[4];


  quaternion_rotation_matrix(player_x*M_PI/180, 0, player_y*M_PI/180, rm);

  glClearColor(1, 1, 1, 1);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glShadeModel(GL_FLAT);
  glEnable(GL_TEXTURE_2D);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  glDepthFunc(GL_LESS);
  glFrontFace(GL_CCW);
  glEnable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);

  glGetIntegerv(GL_VIEWPORT, vp);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45, (double)vp[2]/(double)vp[3], 0.1, 1000);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(distance, 0, 0, 0, 0, 0, 0, 0, 1);
  glPushMatrix();

  glMultMatrixf(rm);

  _render_mesh(meshes[view]);

  glDisable(GL_TEXTURE_2D);
  glPopMatrix();

  SDL_GL_SwapBuffers();

  return SUCCEEDED;
}


int _render_mesh(struct mesh2 *s) {

  struct txt_object *tx;
  int i, n, m, t, id;
  float *uv, *vn, *ve;


  if (s == NULL)
    return FAILED;

  if (view_mode == VIEW_MODE_LINES) {
    /* render the outlines of the triangles */
    glColor3f(0, 0, 0);
    glDisable(GL_TEXTURE_2D);

    glBegin(GL_LINES);
    for (i = 0; i < s->triangles; i++) {
      glVertex3fv(&s->vertex_list[i*3*3 + 0]);
      glVertex3fv(&s->vertex_list[i*3*3 + 3]);
      glVertex3fv(&s->vertex_list[i*3*3 + 3]);
      glVertex3fv(&s->vertex_list[i*3*3 + 6]);
      glVertex3fv(&s->vertex_list[i*3*3 + 6]);
      glVertex3fv(&s->vertex_list[i*3*3 + 0]);
    }
    glEnd();
  }
  else {
    id = -1;
    glColor3f(1, 1, 1);

    uv = s->uv_list;
    vn = s->vertex_normal_list;
    ve = s->vertex_list;

    for (i = 0; i < s->objects; i++) {
      t = s->object_data[i*2 + 0];
      n = s->object_data[i*2 + 1];

      /* txt not loaded yet? */
      if (id != t) {
	if (txt_get(t, &tx) == SUCCEEDED) {
	  glBindTexture(GL_TEXTURE_2D, tx->ogl_name);
	  /* do we have translucent triangles? */
	  if (tx->status & TXT_STATUS_ALPHA) {
	    glEnable(GL_BLEND);
	    glDisable(GL_CULL_FACE);
	    glColor4f(1, 1, 1, tx->alpha);
	  }
	  else {
	    glDisable(GL_BLEND);
	    glEnable(GL_CULL_FACE);
	  }
	  id = t;
	}
      }

      /* draw the triangles */
      glBegin(GL_TRIANGLES);
      for (m = 0; m < n; m++) {
	glNormal3fv(vn);
	glTexCoord2fv(uv);
	glVertex3fv(ve);
	vn += 3;
	uv += 2;
	ve += 3;
	glNormal3fv(vn);
	glTexCoord2fv(uv);
	glVertex3fv(ve);
	vn += 3;
	uv += 2;
	ve += 3;
	glNormal3fv(vn);
	glTexCoord2fv(uv);
	glVertex3fv(ve);
	vn += 3;
	uv += 2;
	ve += 3;
      }
      glEnd();
    }
  }

  return SUCCEEDED;
}
